"""
Enterprise Logging Extension Module.

Provides structured logging, log aggregation, correlation IDs,
and advanced logging patterns for agent operations.

Example:
    # Structured logging
    logger = get_structured_logger("agent")
    logger.info("Processing request", request_id=req_id, user=user)
    
    # Context logging
    with log_context(request_id="123", user="alice"):
        logger.info("Handled request")  # Automatically includes context
    
    # Decorators
    @log_calls(level="INFO")
    async def process_request():
        ...
"""

from __future__ import annotations

import asyncio
import contextvars
import json
import logging
import sys
import time
import traceback
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, asdict
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    TextIO,
    TypeVar,
    Union,
)
from datetime import datetime
from functools import wraps
from contextlib import contextmanager
from enum import Enum

T = TypeVar('T')

# Context variable for log context
_log_context: contextvars.ContextVar[Dict[str, Any]] = contextvars.ContextVar(
    'log_context',
    default={},
)

# Context variable for correlation ID
_correlation_id: contextvars.ContextVar[Optional[str]] = contextvars.ContextVar(
    'correlation_id',
    default=None,
)


class LogLevel(str, Enum):
    """Log levels."""
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"
    
    def to_int(self) -> int:
        """Convert to logging level int."""
        return {
            LogLevel.DEBUG: logging.DEBUG,
            LogLevel.INFO: logging.INFO,
            LogLevel.WARNING: logging.WARNING,
            LogLevel.ERROR: logging.ERROR,
            LogLevel.CRITICAL: logging.CRITICAL,
        }[self]


@dataclass
class LogRecord:
    """A structured log record."""
    timestamp: str
    level: str
    logger: str
    message: str
    correlation_id: Optional[str] = None
    context: Dict[str, Any] = field(default_factory=dict)
    exception: Optional[str] = None
    duration_ms: Optional[float] = None
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary."""
        result = asdict(self)
        # Remove None values
        return {k: v for k, v in result.items() if v is not None}
    
    def to_json(self) -> str:
        """Convert to JSON string."""
        return json.dumps(self.to_dict())


class LogFormatter(ABC):
    """Abstract log formatter."""
    
    @abstractmethod
    def format(self, record: LogRecord) -> str:
        """Format a log record."""
        pass


class JSONFormatter(LogFormatter):
    """JSON log formatter."""
    
    def __init__(self, indent: Optional[int] = None):
        self.indent = indent
    
    def format(self, record: LogRecord) -> str:
        """Format as JSON."""
        return json.dumps(record.to_dict(), indent=self.indent, default=str)


class TextFormatter(LogFormatter):
    """Human-readable text formatter."""
    
    def __init__(
        self,
        format_string: str = "{timestamp} [{level}] {logger}: {message}",
        include_context: bool = True,
    ):
        self.format_string = format_string
        self.include_context = include_context
    
    def format(self, record: LogRecord) -> str:
        """Format as text."""
        base = self.format_string.format(**record.to_dict())
        
        if self.include_context and record.context:
            context_str = " ".join(f"{k}={v}" for k, v in record.context.items())
            base = f"{base} [{context_str}]"
        
        if record.exception:
            base = f"{base}\n{record.exception}"
        
        return base


class ColorFormatter(TextFormatter):
    """Colored text formatter for terminals."""
    
    COLORS = {
        "DEBUG": "\033[36m",     # Cyan
        "INFO": "\033[32m",      # Green
        "WARNING": "\033[33m",   # Yellow
        "ERROR": "\033[31m",     # Red
        "CRITICAL": "\033[35m",  # Magenta
    }
    RESET = "\033[0m"
    
    def format(self, record: LogRecord) -> str:
        """Format with colors."""
        base = super().format(record)
        color = self.COLORS.get(record.level, "")
        return f"{color}{base}{self.RESET}"


class LogSink(ABC):
    """Abstract log sink/output."""
    
    @abstractmethod
    def write(self, record: LogRecord) -> None:
        """Write a log record."""
        pass
    
    def flush(self) -> None:
        """Flush the sink."""
        pass
    
    def close(self) -> None:
        """Close the sink."""
        pass


class ConsoleSink(LogSink):
    """Console log sink."""
    
    def __init__(
        self,
        formatter: Optional[LogFormatter] = None,
        stream: TextIO = sys.stderr,
    ):
        self.formatter = formatter or TextFormatter()
        self.stream = stream
    
    def write(self, record: LogRecord) -> None:
        """Write to console."""
        self.stream.write(self.formatter.format(record) + "\n")
        self.stream.flush()


class FileSink(LogSink):
    """File log sink with rotation support."""
    
    def __init__(
        self,
        filepath: str,
        formatter: Optional[LogFormatter] = None,
        max_size_mb: float = 10.0,
        backup_count: int = 5,
    ):
        self.filepath = filepath
        self.formatter = formatter or JSONFormatter()
        self.max_size_mb = max_size_mb
        self.backup_count = backup_count
        self._file: Optional[TextIO] = None
        self._current_size = 0
    
    def _open(self) -> None:
        """Open the log file."""
        self._file = open(self.filepath, "a", encoding="utf-8")
    
    def _should_rotate(self) -> bool:
        """Check if rotation is needed."""
        return self._current_size >= self.max_size_mb * 1024 * 1024
    
    def _rotate(self) -> None:
        """Rotate log files."""
        if self._file:
            self._file.close()
        
        import os
        import shutil
        
        # Rotate existing files
        for i in range(self.backup_count - 1, 0, -1):
            src = f"{self.filepath}.{i}"
            dst = f"{self.filepath}.{i + 1}"
            if os.path.exists(src):
                shutil.move(src, dst)
        
        # Move current file
        if os.path.exists(self.filepath):
            shutil.move(self.filepath, f"{self.filepath}.1")
        
        self._current_size = 0
        self._open()
    
    def write(self, record: LogRecord) -> None:
        """Write to file."""
        if self._file is None:
            self._open()
        
        if self._should_rotate():
            self._rotate()
        
        line = self.formatter.format(record) + "\n"
        self._file.write(line)
        self._current_size += len(line.encode())
    
    def flush(self) -> None:
        """Flush the file."""
        if self._file:
            self._file.flush()
    
    def close(self) -> None:
        """Close the file."""
        if self._file:
            self._file.close()
            self._file = None


class MemorySink(LogSink):
    """In-memory log sink for testing."""
    
    def __init__(self, max_records: int = 1000):
        self.records: List[LogRecord] = []
        self.max_records = max_records
    
    def write(self, record: LogRecord) -> None:
        """Store in memory."""
        self.records.append(record)
        if len(self.records) > self.max_records:
            self.records = self.records[-self.max_records:]
    
    def clear(self) -> None:
        """Clear records."""
        self.records.clear()
    
    def search(
        self,
        level: Optional[str] = None,
        message_contains: Optional[str] = None,
        context_key: Optional[str] = None,
    ) -> List[LogRecord]:
        """Search records."""
        results = self.records
        
        if level:
            results = [r for r in results if r.level == level]
        
        if message_contains:
            results = [r for r in results if message_contains in r.message]
        
        if context_key:
            results = [r for r in results if context_key in r.context]
        
        return results


class AggregateSink(LogSink):
    """Aggregate multiple sinks."""
    
    def __init__(self, sinks: Optional[List[LogSink]] = None):
        self.sinks = sinks or []
    
    def add(self, sink: LogSink) -> 'AggregateSink':
        """Add a sink."""
        self.sinks.append(sink)
        return self
    
    def write(self, record: LogRecord) -> None:
        """Write to all sinks."""
        for sink in self.sinks:
            try:
                sink.write(record)
            except Exception:
                pass  # Don't fail on sink errors
    
    def flush(self) -> None:
        """Flush all sinks."""
        for sink in self.sinks:
            sink.flush()
    
    def close(self) -> None:
        """Close all sinks."""
        for sink in self.sinks:
            sink.close()


class FilteredSink(LogSink):
    """Sink with filtering."""
    
    def __init__(
        self,
        sink: LogSink,
        min_level: LogLevel = LogLevel.DEBUG,
        include_loggers: Optional[List[str]] = None,
        exclude_loggers: Optional[List[str]] = None,
    ):
        self.sink = sink
        self.min_level = min_level
        self.include_loggers = include_loggers
        self.exclude_loggers = exclude_loggers
    
    def _should_log(self, record: LogRecord) -> bool:
        """Check if record should be logged."""
        # Check level
        level = LogLevel(record.level)
        if level.to_int() < self.min_level.to_int():
            return False
        
        # Check logger inclusion
        if self.include_loggers:
            if not any(record.logger.startswith(l) for l in self.include_loggers):
                return False
        
        # Check logger exclusion
        if self.exclude_loggers:
            if any(record.logger.startswith(l) for l in self.exclude_loggers):
                return False
        
        return True
    
    def write(self, record: LogRecord) -> None:
        """Write if passes filter."""
        if self._should_log(record):
            self.sink.write(record)


class StructuredLogger:
    """
    Structured logger with context support.
    """
    
    def __init__(
        self,
        name: str,
        sink: Optional[LogSink] = None,
        min_level: LogLevel = LogLevel.DEBUG,
    ):
        """
        Initialize structured logger.
        
        Args:
            name: Logger name
            sink: Log sink
            min_level: Minimum log level
        """
        self.name = name
        self.sink = sink or ConsoleSink(TextFormatter())
        self.min_level = min_level
    
    def _create_record(
        self,
        level: LogLevel,
        message: str,
        exc_info: Optional[Exception] = None,
        duration_ms: Optional[float] = None,
        **kwargs: Any,
    ) -> LogRecord:
        """Create a log record."""
        # Get context
        context = {**_log_context.get(), **kwargs}
        
        # Get correlation ID
        correlation_id = _correlation_id.get()
        
        # Format exception
        exception = None
        if exc_info:
            exception = "".join(traceback.format_exception(
                type(exc_info), exc_info, exc_info.__traceback__
            ))
        
        return LogRecord(
            timestamp=datetime.now().isoformat(),
            level=level.value,
            logger=self.name,
            message=message,
            correlation_id=correlation_id,
            context=context,
            exception=exception,
            duration_ms=duration_ms,
        )
    
    def _log(
        self,
        level: LogLevel,
        message: str,
        exc_info: Optional[Exception] = None,
        **kwargs: Any,
    ) -> None:
        """Log a message."""
        if level.to_int() < self.min_level.to_int():
            return
        
        record = self._create_record(level, message, exc_info, **kwargs)
        self.sink.write(record)
    
    def debug(self, message: str, **kwargs: Any) -> None:
        """Log debug message."""
        self._log(LogLevel.DEBUG, message, **kwargs)
    
    def info(self, message: str, **kwargs: Any) -> None:
        """Log info message."""
        self._log(LogLevel.INFO, message, **kwargs)
    
    def warning(self, message: str, **kwargs: Any) -> None:
        """Log warning message."""
        self._log(LogLevel.WARNING, message, **kwargs)
    
    def error(self, message: str, exc_info: Optional[Exception] = None, **kwargs: Any) -> None:
        """Log error message."""
        self._log(LogLevel.ERROR, message, exc_info, **kwargs)
    
    def critical(self, message: str, exc_info: Optional[Exception] = None, **kwargs: Any) -> None:
        """Log critical message."""
        self._log(LogLevel.CRITICAL, message, exc_info, **kwargs)
    
    def exception(self, message: str, exc: Exception, **kwargs: Any) -> None:
        """Log exception."""
        self.error(message, exc_info=exc, **kwargs)
    
    @contextmanager
    def timed(self, operation: str, level: LogLevel = LogLevel.INFO, **kwargs: Any):
        """
        Context manager to log operation duration.
        
        Example:
            with logger.timed("process_request"):
                process()
        """
        start = time.time()
        try:
            yield
        finally:
            duration_ms = (time.time() - start) * 1000
            record = self._create_record(
                level,
                f"{operation} completed",
                duration_ms=duration_ms,
                **kwargs,
            )
            self.sink.write(record)
    
    def child(self, name: str) -> 'StructuredLogger':
        """Create a child logger."""
        return StructuredLogger(
            f"{self.name}.{name}",
            self.sink,
            self.min_level,
        )


# Logger registry
_loggers: Dict[str, StructuredLogger] = {}
_default_sink: Optional[LogSink] = None
_default_level: LogLevel = LogLevel.INFO


def configure_logging(
    sink: Optional[LogSink] = None,
    level: LogLevel = LogLevel.INFO,
) -> None:
    """
    Configure global logging settings.
    
    Args:
        sink: Default sink for new loggers
        level: Default minimum level
    """
    global _default_sink, _default_level
    _default_sink = sink
    _default_level = level


def get_structured_logger(name: str) -> StructuredLogger:
    """
    Get or create a structured logger.
    
    Args:
        name: Logger name
        
    Returns:
        StructuredLogger instance
    """
    if name not in _loggers:
        sink = _default_sink or ConsoleSink(ColorFormatter())
        _loggers[name] = StructuredLogger(name, sink, _default_level)
    
    return _loggers[name]


@contextmanager
def log_context(**kwargs: Any):
    """
    Context manager to add logging context.
    
    Example:
        with log_context(request_id="123", user="alice"):
            logger.info("Processing")  # Includes request_id and user
    """
    current = _log_context.get()
    new_context = {**current, **kwargs}
    token = _log_context.set(new_context)
    try:
        yield
    finally:
        _log_context.reset(token)


@contextmanager
def correlation_context(correlation_id: Optional[str] = None):
    """
    Context manager for correlation ID.
    
    Example:
        with correlation_context("req-123"):
            logger.info("Request received")  # Includes correlation_id
    """
    if correlation_id is None:
        correlation_id = str(uuid.uuid4())
    
    token = _correlation_id.set(correlation_id)
    try:
        yield correlation_id
    finally:
        _correlation_id.reset(token)


def get_correlation_id() -> Optional[str]:
    """Get current correlation ID."""
    return _correlation_id.get()


def set_correlation_id(correlation_id: str) -> None:
    """Set correlation ID."""
    _correlation_id.set(correlation_id)


def log_calls(
    level: Union[str, LogLevel] = LogLevel.INFO,
    include_args: bool = False,
    include_result: bool = False,
    logger_name: Optional[str] = None,
) -> Callable:
    """
    Decorator to log function calls.
    
    Example:
        @log_calls(level="INFO", include_args=True)
        async def process_request(data):
            ...
    """
    if isinstance(level, str):
        level = LogLevel(level)
    
    def decorator(func: Callable) -> Callable:
        log_name = logger_name or func.__module__
        logger = get_structured_logger(log_name)
        
        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            extra = {"function": func.__name__}
            
            if include_args:
                extra["args"] = str(args)[:100]
                extra["kwargs"] = str(kwargs)[:100]
            
            logger._log(level, f"Calling {func.__name__}", **extra)
            
            start = time.time()
            try:
                result = await func(*args, **kwargs)
                
                duration_ms = (time.time() - start) * 1000
                result_extra = {"duration_ms": duration_ms}
                
                if include_result:
                    result_extra["result"] = str(result)[:100]
                
                logger._log(level, f"Completed {func.__name__}", **result_extra)
                return result
                
            except Exception as e:
                duration_ms = (time.time() - start) * 1000
                logger.exception(
                    f"Failed {func.__name__}",
                    e,
                    duration_ms=duration_ms,
                    function=func.__name__,
                )
                raise
        
        @wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            extra = {"function": func.__name__}
            
            if include_args:
                extra["args"] = str(args)[:100]
                extra["kwargs"] = str(kwargs)[:100]
            
            logger._log(level, f"Calling {func.__name__}", **extra)
            
            start = time.time()
            try:
                result = func(*args, **kwargs)
                
                duration_ms = (time.time() - start) * 1000
                result_extra = {"duration_ms": duration_ms}
                
                if include_result:
                    result_extra["result"] = str(result)[:100]
                
                logger._log(level, f"Completed {func.__name__}", **result_extra)
                return result
                
            except Exception as e:
                duration_ms = (time.time() - start) * 1000
                logger.exception(
                    f"Failed {func.__name__}",
                    e,
                    duration_ms=duration_ms,
                    function=func.__name__,
                )
                raise
        
        return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper
    
    return decorator


def log_errors(
    logger_name: Optional[str] = None,
    reraise: bool = True,
) -> Callable:
    """
    Decorator to log exceptions.
    
    Example:
        @log_errors()
        async def risky_operation():
            ...
    """
    def decorator(func: Callable) -> Callable:
        log_name = logger_name or func.__module__
        logger = get_structured_logger(log_name)
        
        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                logger.exception(
                    f"Exception in {func.__name__}",
                    e,
                    function=func.__name__,
                )
                if reraise:
                    raise
                return None
        
        @wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return func(*args, **kwargs)
            except Exception as e:
                logger.exception(
                    f"Exception in {func.__name__}",
                    e,
                    function=func.__name__,
                )
                if reraise:
                    raise
                return None
        
        return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper
    
    return decorator


class StandardLibraryAdapter(logging.Handler):
    """
    Adapter to use structured logger with standard library logging.
    """
    
    def __init__(self, sink: LogSink, level: int = logging.DEBUG):
        super().__init__(level)
        self.sink = sink
    
    def emit(self, record: logging.LogRecord) -> None:
        """Emit a standard library log record."""
        log_record = LogRecord(
            timestamp=datetime.fromtimestamp(record.created).isoformat(),
            level=record.levelname,
            logger=record.name,
            message=record.getMessage(),
            correlation_id=_correlation_id.get(),
            context=_log_context.get(),
            exception=self.format_exception(record) if record.exc_info else None,
        )
        self.sink.write(log_record)
    
    def format_exception(self, record: logging.LogRecord) -> str:
        """Format exception info."""
        if record.exc_info:
            return "".join(traceback.format_exception(*record.exc_info))
        return ""


def setup_stdlib_integration(sink: Optional[LogSink] = None) -> None:
    """
    Integrate structured logging with standard library.
    
    This redirects all standard library logging to the structured sink.
    """
    sink = sink or ConsoleSink(JSONFormatter())
    handler = StandardLibraryAdapter(sink)
    
    root = logging.getLogger()
    root.handlers.clear()
    root.addHandler(handler)
    root.setLevel(logging.DEBUG)


__all__ = [
    # Enums
    "LogLevel",
    # Data classes
    "LogRecord",
    # Formatters
    "LogFormatter",
    "JSONFormatter",
    "TextFormatter",
    "ColorFormatter",
    # Sinks
    "LogSink",
    "ConsoleSink",
    "FileSink",
    "MemorySink",
    "AggregateSink",
    "FilteredSink",
    # Logger
    "StructuredLogger",
    # Configuration
    "configure_logging",
    "get_structured_logger",
    # Context
    "log_context",
    "correlation_context",
    "get_correlation_id",
    "set_correlation_id",
    # Decorators
    "log_calls",
    "log_errors",
    # Integration
    "StandardLibraryAdapter",
    "setup_stdlib_integration",
]



"""
Enterprise Blueprints - Pre-built agent configurations for common use cases.

Blueprints provide ready-to-use agent configurations that can be
instantiated with minimal code.

Usage:
    >>> from agenticaiframework.enterprise.blueprints import RequirementsAgent
    >>> 
    >>> agent = RequirementsAgent()
    >>> result = await agent.analyze("Build an e-commerce platform")
"""

import asyncio
import logging
import os
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Type

logger = logging.getLogger(__name__)


# =============================================================================
# Blueprint Base
# =============================================================================

@dataclass
class BlueprintConfig:
    """Configuration for blueprint agents."""
    name: str
    role: str
    capabilities: List[str] = field(default_factory=list)
    model: str = "gpt-4o"
    provider: str = "azure"
    temperature: float = 0.7
    max_tokens: int = 4096
    system_prompt: str = ""


class BlueprintAgent:
    """
    Base class for blueprint agents.
    
    Provides a consistent interface for all pre-built agents.
    """
    
    config: BlueprintConfig
    
    def __init__(
        self,
        *,
        model: Optional[str] = None,
        provider: Optional[str] = None,
        **kwargs,
    ):
        self.model = model or self.config.model
        self.provider = provider or self.config.provider
        self._client = None
        self._history: List[Dict] = []
        self._kwargs = kwargs
    
    def _get_client(self):
        """Get or create LLM client."""
        if self._client is None:
            if self.provider == "azure":
                try:
                    from openai import AzureOpenAI
                    self._client = AzureOpenAI(
                        api_key=os.getenv("AZURE_OPENAI_API_KEY"),
                        azure_endpoint=os.getenv("AZURE_OPENAI_ENDPOINT"),
                        api_version="2024-02-01",
                    )
                except ImportError:
                    raise ImportError("openai is required. Install with: pip install openai")
        return self._client
    
    async def _call_llm(self, prompt: str, system: Optional[str] = None) -> str:
        """Call LLM with the prompt."""
        client = self._get_client()
        
        messages = []
        if system or self.config.system_prompt:
            messages.append({
                "role": "system",
                "content": system or self.config.system_prompt,
            })
        messages.append({"role": "user", "content": prompt})
        
        params = {
            "model": self.model,
            "messages": messages,
        }
        
        # Handle model-specific parameters
        if "gpt-5" in self.model.lower():
            params["max_completion_tokens"] = self.config.max_tokens
            params["temperature"] = 1
        else:
            params["max_tokens"] = self.config.max_tokens
            params["temperature"] = self.config.temperature
        
        response = await asyncio.to_thread(
            lambda: client.chat.completions.create(**params)
        )
        
        return response.choices[0].message.content
    
    async def invoke(self, prompt: str) -> str:
        """
        Invoke the agent with a prompt.
        
        Args:
            prompt: User prompt
            
        Returns:
            Agent response
        """
        return await self._call_llm(prompt)
    
    def invoke_sync(self, prompt: str) -> str:
        """Synchronous version of invoke."""
        return asyncio.run(self.invoke(prompt))


# =============================================================================
# SDLC Blueprint Agents
# =============================================================================

class RequirementsAgent(BlueprintAgent):
    """
    Pre-built Requirements Analyst agent.
    
    Analyzes project descriptions and creates structured requirements
    including functional requirements, non-functional requirements,
    and user stories.
    
    Usage:
        >>> agent = RequirementsAgent()
        >>> result = await agent.analyze("Build an e-commerce platform")
    """
    
    config = BlueprintConfig(
        name="RequirementsAgent",
        role="Requirements Analyst",
        capabilities=["requirements-analysis", "user-stories", "use-cases"],
        system_prompt="""You are an expert Requirements Analyst with deep experience in software development.
Your task is to analyze project descriptions and create comprehensive, structured requirements.

Always include:
1. Functional Requirements (FR-001, FR-002, etc.)
2. Non-Functional Requirements (NFR-001, etc.)
3. User Stories in format: "As a [user], I want [feature] so that [benefit]"
4. Acceptance Criteria for each requirement
5. Priority levels (Must Have, Should Have, Nice to Have)
6. Dependencies between requirements

Format your output as clean, structured markdown.""",
    )
    
    async def analyze(self, description: str) -> str:
        """Analyze a project description and generate requirements."""
        prompt = f"""Analyze the following project and create detailed software requirements:

{description}

Provide comprehensive requirements in structured markdown format."""
        return await self.invoke(prompt)


class DesignAgent(BlueprintAgent):
    """
    Pre-built Software Architect agent.
    
    Creates technical designs including architecture, database schemas,
    and API specifications.
    
    Usage:
        >>> agent = DesignAgent()
        >>> result = await agent.design(requirements)
    """
    
    config = BlueprintConfig(
        name="DesignAgent",
        role="Software Architect",
        capabilities=["architecture", "database-design", "api-design"],
        system_prompt="""You are an expert Software Architect with experience in scalable, maintainable systems.
Your task is to create comprehensive technical designs based on requirements.

Always include:
1. System Architecture with component diagram description
2. Database Design (tables, relationships, indexes)
3. API Design (endpoints, methods, request/response schemas)
4. Technology Stack recommendations with justification
5. Security considerations
6. Scalability and performance strategies
7. Error handling approach

Format your output as clean, structured markdown with code examples where appropriate.""",
    )
    
    async def design(self, requirements: str) -> str:
        """Create a technical design based on requirements."""
        prompt = f"""Based on the following requirements, create a comprehensive technical design:

{requirements}

Provide a detailed technical design document in markdown format."""
        return await self.invoke(prompt)


class DevelopmentAgent(BlueprintAgent):
    """
    Pre-built Senior Developer agent.
    
    Generates production-quality code based on technical designs.
    
    Usage:
        >>> agent = DevelopmentAgent()
        >>> result = await agent.implement(design)
    """
    
    config = BlueprintConfig(
        name="DevelopmentAgent",
        role="Senior Software Developer",
        capabilities=["code-generation", "refactoring", "best-practices"],
        max_tokens=8192,
        system_prompt="""You are a Senior Software Developer with expertise in clean code and best practices.
Your task is to implement production-quality code based on technical designs.

Always follow:
1. Clean Code principles (SOLID, DRY, KISS)
2. Comprehensive error handling
3. Logging and monitoring hooks
4. Configuration management
5. Security best practices
6. Performance considerations
7. Proper documentation

Provide complete, runnable code with comments explaining complex logic.""",
    )
    
    async def implement(self, design: str) -> str:
        """Implement code based on a technical design."""
        prompt = f"""Based on the following technical design, implement production-quality code:

{design}

Provide complete, well-documented code with proper error handling."""
        return await self.invoke(prompt)


class TestingAgent(BlueprintAgent):
    """
    Pre-built QA Engineer agent.
    
    Creates test strategies, test cases, and test automation code.
    
    Usage:
        >>> agent = TestingAgent()
        >>> result = await agent.create_tests(code)
    """
    
    config = BlueprintConfig(
        name="TestingAgent",
        role="QA Engineer",
        capabilities=["test-strategy", "test-cases", "automation"],
        system_prompt="""You are an expert QA Engineer with deep experience in testing methodologies.
Your task is to create comprehensive testing strategies and test cases.

Always include:
1. Test Strategy document
2. Unit test cases with expected results
3. Integration test scenarios
4. Edge case coverage
5. Performance test considerations
6. Security test checklist
7. Test data requirements

Format as structured markdown with actual test code where appropriate.""",
    )
    
    async def create_tests(self, implementation: str) -> str:
        """Create test strategy and test cases."""
        prompt = f"""Based on the following implementation, create a comprehensive test strategy:

{implementation}

Provide test strategy and test cases in markdown format with test code."""
        return await self.invoke(prompt)


class SecurityAgent(BlueprintAgent):
    """
    Pre-built Security Analyst agent.
    
    Performs security analysis and provides recommendations.
    
    Usage:
        >>> agent = SecurityAgent()
        >>> result = await agent.analyze(code)
    """
    
    config = BlueprintConfig(
        name="SecurityAgent",
        role="Security Analyst",
        capabilities=["threat-modeling", "vulnerability-analysis", "security-review"],
        system_prompt="""You are an expert Security Analyst with experience in application security.
Your task is to perform security analysis and provide actionable recommendations.

Always include:
1. Threat Model (STRIDE analysis)
2. Vulnerability Assessment
3. Security Requirements
4. Authentication/Authorization review
5. Data protection recommendations
6. Input validation analysis
7. Compliance considerations (GDPR, SOC2, etc.)

Format as structured markdown with severity ratings and remediation steps.""",
    )
    
    async def analyze(self, implementation: str) -> str:
        """Perform security analysis."""
        prompt = f"""Perform a comprehensive security analysis on the following:

{implementation}

Provide a security report with vulnerabilities and recommendations."""
        return await self.invoke(prompt)


class DeploymentAgent(BlueprintAgent):
    """
    Pre-built DevOps Engineer agent.
    
    Creates deployment configurations and infrastructure code.
    
    Usage:
        >>> agent = DeploymentAgent()
        >>> result = await agent.create_deployment(project)
    """
    
    config = BlueprintConfig(
        name="DeploymentAgent",
        role="DevOps Engineer",
        capabilities=["containerization", "ci-cd", "infrastructure"],
        system_prompt="""You are an expert DevOps Engineer with experience in cloud deployments.
Your task is to create deployment configurations and infrastructure code.

Always include:
1. Dockerfile (multi-stage, optimized)
2. docker-compose.yml
3. Kubernetes manifests (if applicable)
4. CI/CD pipeline configuration (GitHub Actions)
5. Infrastructure as Code (Terraform/Bicep)
6. Environment configuration
7. Monitoring and alerting setup

Provide production-ready configuration files with comments.""",
    )
    
    async def create_deployment(self, project: str) -> str:
        """Create deployment configuration."""
        prompt = f"""Create comprehensive deployment configuration for:

{project}

Provide all necessary deployment files with comments."""
        return await self.invoke(prompt)


class DocumentationAgent(BlueprintAgent):
    """
    Pre-built Technical Writer agent.
    
    Creates comprehensive project documentation.
    
    Usage:
        >>> agent = DocumentationAgent()
        >>> result = await agent.document(project)
    """
    
    config = BlueprintConfig(
        name="DocumentationAgent",
        role="Technical Writer",
        capabilities=["documentation", "api-docs", "user-guides"],
        system_prompt="""You are an expert Technical Writer with experience in software documentation.
Your task is to create comprehensive, user-friendly documentation.

Always include:
1. README.md with quick start guide
2. API documentation with examples
3. Architecture documentation
4. Configuration guide
5. Troubleshooting section
6. Contributing guidelines

Format as clean, accessible markdown with code examples.""",
    )
    
    async def document(self, project: str) -> str:
        """Create project documentation."""
        prompt = f"""Create comprehensive documentation for:

{project}

Provide complete documentation in markdown format."""
        return await self.invoke(prompt)


class ReviewAgent(BlueprintAgent):
    """
    Pre-built Code Reviewer agent.
    
    Reviews code and provides feedback.
    
    Usage:
        >>> agent = ReviewAgent()
        >>> result = await agent.review(code)
    """
    
    config = BlueprintConfig(
        name="ReviewAgent",
        role="Senior Code Reviewer",
        capabilities=["code-review", "best-practices", "quality-assurance"],
        system_prompt="""You are an expert Code Reviewer with deep experience in software quality.
Your task is to review code and provide constructive feedback.

Always check for:
1. Code quality and readability
2. Best practices compliance
3. Security vulnerabilities
4. Performance issues
5. Error handling completeness
6. Test coverage
7. Documentation quality

Format as structured markdown with severity ratings and specific recommendations.""",
    )
    
    async def review(self, code: str) -> str:
        """Review code and provide feedback."""
        prompt = f"""Review the following code and provide detailed feedback:

{code}

Provide a structured review with recommendations."""
        return await self.invoke(prompt)


# =============================================================================
# Utility Blueprint Agents
# =============================================================================

class AnalystAgent(BlueprintAgent):
    """General data analyst agent."""
    
    config = BlueprintConfig(
        name="AnalystAgent",
        role="Data Analyst",
        capabilities=["data-analysis", "visualization", "insights"],
        system_prompt="""You are a Data Analyst expert at extracting insights from data.
Provide clear, actionable analysis with visualizations where helpful.""",
    )
    
    async def analyze(self, data: str) -> str:
        """Analyze data and provide insights."""
        return await self.invoke(f"Analyze the following data:\n\n{data}")


class WriterAgent(BlueprintAgent):
    """Creative writing agent."""
    
    config = BlueprintConfig(
        name="WriterAgent",
        role="Content Writer",
        capabilities=["writing", "editing", "content-creation"],
        temperature=0.8,
        system_prompt="""You are a creative Content Writer skilled at engaging content.
Write clear, compelling content tailored to the audience.""",
    )
    
    async def write(self, brief: str) -> str:
        """Write content based on a brief."""
        return await self.invoke(f"Write content based on:\n\n{brief}")


class ResearcherAgent(BlueprintAgent):
    """Research assistant agent."""
    
    config = BlueprintConfig(
        name="ResearcherAgent",
        role="Research Assistant",
        capabilities=["research", "summarization", "synthesis"],
        system_prompt="""You are a Research Assistant skilled at finding and synthesizing information.
Provide well-organized, cited research with clear conclusions.""",
    )
    
    async def research(self, topic: str) -> str:
        """Research a topic."""
        return await self.invoke(f"Research the following topic:\n\n{topic}")


# =============================================================================
# Blueprint Registry
# =============================================================================

BLUEPRINT_REGISTRY: Dict[str, Type[BlueprintAgent]] = {
    # SDLC Agents
    "requirements": RequirementsAgent,
    "design": DesignAgent,
    "development": DevelopmentAgent,
    "testing": TestingAgent,
    "security": SecurityAgent,
    "deployment": DeploymentAgent,
    "documentation": DocumentationAgent,
    "review": ReviewAgent,
    # Utility Agents
    "analyst": AnalystAgent,
    "writer": WriterAgent,
    "researcher": ResearcherAgent,
}


def get_blueprint(name: str, **kwargs) -> BlueprintAgent:
    """
    Get a blueprint agent by name.
    
    Args:
        name: Blueprint name
        **kwargs: Configuration overrides
        
    Returns:
        Blueprint agent instance
    """
    agent_class = BLUEPRINT_REGISTRY.get(name.lower())
    
    if agent_class is None:
        available = ", ".join(BLUEPRINT_REGISTRY.keys())
        raise ValueError(f"Unknown blueprint: {name}. Available: {available}")
    
    return agent_class(**kwargs)


def list_blueprints() -> List[str]:
    """List available blueprints."""
    return list(BLUEPRINT_REGISTRY.keys())



"""
Enterprise Schema Module.

Provides schema validation, type checking, and data
structure verification capabilities.

Example:
    # Define schema
    UserSchema = Schema({
        "name": String(required=True, min_length=1),
        "age": Integer(min=0, max=150),
        "email": Email(required=True),
        "tags": Array(String()),
    })
    
    # Validate data
    result = UserSchema.validate(data)
    if result.is_valid:
        process(result.data)
    else:
        handle_errors(result.errors)
    
    # With decorator
    @validate_schema(UserSchema)
    def create_user(data: dict) -> User:
        return User(**data)
    
    # Schema from dataclass
    @dataclass_schema
    class User:
        name: str
        age: int = 0
"""

from __future__ import annotations

import re
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, fields as dataclass_fields
from datetime import datetime, date
from enum import Enum
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Pattern,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_type_hints,
)

logger = logging.getLogger(__name__)

T = TypeVar('T')


class SchemaError(Exception):
    """Schema error."""
    pass


class ValidationError(SchemaError):
    """Validation error."""
    
    def __init__(
        self,
        message: str,
        path: Optional[str] = None,
        value: Any = None,
    ):
        self.message = message
        self.path = path
        self.value = value
        super().__init__(message)


class SchemaType(str, Enum):
    """Schema types."""
    STRING = "string"
    INTEGER = "integer"
    FLOAT = "float"
    BOOLEAN = "boolean"
    ARRAY = "array"
    OBJECT = "object"
    ANY = "any"
    NULL = "null"
    DATE = "date"
    DATETIME = "datetime"
    EMAIL = "email"
    URL = "url"
    UUID = "uuid"
    ENUM = "enum"


@dataclass
class ValidationResult:
    """Validation result."""
    is_valid: bool
    data: Any = None
    errors: List[ValidationError] = field(default_factory=list)
    warnings: List[str] = field(default_factory=list)
    
    def merge(self, other: "ValidationResult") -> "ValidationResult":
        """Merge with another result."""
        return ValidationResult(
            is_valid=self.is_valid and other.is_valid,
            data=other.data if other.is_valid else self.data,
            errors=self.errors + other.errors,
            warnings=self.warnings + other.warnings,
        )


@dataclass
class FieldInfo:
    """Field information."""
    name: str
    type: "SchemaField"
    required: bool = True
    default: Any = None
    description: Optional[str] = None


class SchemaField(ABC):
    """Abstract schema field."""
    
    def __init__(
        self,
        required: bool = False,
        default: Any = None,
        nullable: bool = False,
        description: Optional[str] = None,
    ):
        self.required = required
        self.default = default
        self.nullable = nullable
        self.description = description
    
    @abstractmethod
    def validate(
        self,
        value: Any,
        path: str = "",
    ) -> ValidationResult:
        """Validate value."""
        pass
    
    def _check_null(
        self,
        value: Any,
        path: str,
    ) -> Optional[ValidationResult]:
        """Check for null value."""
        if value is None:
            if self.nullable:
                return ValidationResult(is_valid=True, data=None)
            
            if self.required:
                return ValidationResult(
                    is_valid=False,
                    errors=[ValidationError(
                        f"Required field is null",
                        path=path,
                        value=value,
                    )],
                )
            
            return ValidationResult(is_valid=True, data=self.default)
        
        return None


class String(SchemaField):
    """String schema field."""
    
    def __init__(
        self,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        pattern: Optional[str] = None,
        choices: Optional[List[str]] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.min_length = min_length
        self.max_length = max_length
        self.pattern = re.compile(pattern) if pattern else None
        self.choices = choices
    
    def validate(
        self,
        value: Any,
        path: str = "",
    ) -> ValidationResult:
        """Validate string."""
        null_result = self._check_null(value, path)
        if null_result:
            return null_result
        
        if not isinstance(value, str):
            return ValidationResult(
                is_valid=False,
                errors=[ValidationError(
                    f"Expected string, got {type(value).__name__}",
                    path=path,
                    value=value,
                )],
            )
        
        errors = []
        
        if self.min_length is not None and len(value) < self.min_length:
            errors.append(ValidationError(
                f"String length {len(value)} is less than minimum {self.min_length}",
                path=path,
                value=value,
            ))
        
        if self.max_length is not None and len(value) > self.max_length:
            errors.append(ValidationError(
                f"String length {len(value)} exceeds maximum {self.max_length}",
                path=path,
                value=value,
            ))
        
        if self.pattern and not self.pattern.match(value):
            errors.append(ValidationError(
                f"String does not match pattern",
                path=path,
                value=value,
            ))
        
        if self.choices and value not in self.choices:
            errors.append(ValidationError(
                f"Value not in allowed choices: {self.choices}",
                path=path,
                value=value,
            ))
        
        return ValidationResult(
            is_valid=len(errors) == 0,
            data=value,
            errors=errors,
        )


class Integer(SchemaField):
    """Integer schema field."""
    
    def __init__(
        self,
        min: Optional[int] = None,
        max: Optional[int] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.min = min
        self.max = max
    
    def validate(
        self,
        value: Any,
        path: str = "",
    ) -> ValidationResult:
        """Validate integer."""
        null_result = self._check_null(value, path)
        if null_result:
            return null_result
        
        # Allow string conversion
        if isinstance(value, str):
            try:
                value = int(value)
            except ValueError:
                return ValidationResult(
                    is_valid=False,
                    errors=[ValidationError(
                        f"Cannot convert '{value}' to integer",
                        path=path,
                        value=value,
                    )],
                )
        
        if not isinstance(value, int) or isinstance(value, bool):
            return ValidationResult(
                is_valid=False,
                errors=[ValidationError(
                    f"Expected integer, got {type(value).__name__}",
                    path=path,
                    value=value,
                )],
            )
        
        errors = []
        
        if self.min is not None and value < self.min:
            errors.append(ValidationError(
                f"Value {value} is less than minimum {self.min}",
                path=path,
                value=value,
            ))
        
        if self.max is not None and value > self.max:
            errors.append(ValidationError(
                f"Value {value} exceeds maximum {self.max}",
                path=path,
                value=value,
            ))
        
        return ValidationResult(
            is_valid=len(errors) == 0,
            data=value,
            errors=errors,
        )


class Float(SchemaField):
    """Float schema field."""
    
    def __init__(
        self,
        min: Optional[float] = None,
        max: Optional[float] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.min = min
        self.max = max
    
    def validate(
        self,
        value: Any,
        path: str = "",
    ) -> ValidationResult:
        """Validate float."""
        null_result = self._check_null(value, path)
        if null_result:
            return null_result
        
        # Allow string/int conversion
        if isinstance(value, (str, int)) and not isinstance(value, bool):
            try:
                value = float(value)
            except ValueError:
                return ValidationResult(
                    is_valid=False,
                    errors=[ValidationError(
                        f"Cannot convert '{value}' to float",
                        path=path,
                        value=value,
                    )],
                )
        
        if not isinstance(value, (int, float)) or isinstance(value, bool):
            return ValidationResult(
                is_valid=False,
                errors=[ValidationError(
                    f"Expected float, got {type(value).__name__}",
                    path=path,
                    value=value,
                )],
            )
        
        errors = []
        
        if self.min is not None and value < self.min:
            errors.append(ValidationError(
                f"Value {value} is less than minimum {self.min}",
                path=path,
                value=value,
            ))
        
        if self.max is not None and value > self.max:
            errors.append(ValidationError(
                f"Value {value} exceeds maximum {self.max}",
                path=path,
                value=value,
            ))
        
        return ValidationResult(
            is_valid=len(errors) == 0,
            data=float(value),
            errors=errors,
        )


class Boolean(SchemaField):
    """Boolean schema field."""
    
    def validate(
        self,
        value: Any,
        path: str = "",
    ) -> ValidationResult:
        """Validate boolean."""
        null_result = self._check_null(value, path)
        if null_result:
            return null_result
        
        # Allow string conversion
        if isinstance(value, str):
            if value.lower() in ('true', '1', 'yes', 'on'):
                value = True
            elif value.lower() in ('false', '0', 'no', 'off'):
                value = False
            else:
                return ValidationResult(
                    is_valid=False,
                    errors=[ValidationError(
                        f"Cannot convert '{value}' to boolean",
                        path=path,
                        value=value,
                    )],
                )
        
        if not isinstance(value, bool):
            return ValidationResult(
                is_valid=False,
                errors=[ValidationError(
                    f"Expected boolean, got {type(value).__name__}",
                    path=path,
                    value=value,
                )],
            )
        
        return ValidationResult(is_valid=True, data=value)


class Array(SchemaField):
    """Array schema field."""
    
    def __init__(
        self,
        items: Optional[SchemaField] = None,
        min_items: Optional[int] = None,
        max_items: Optional[int] = None,
        unique: bool = False,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.items = items
        self.min_items = min_items
        self.max_items = max_items
        self.unique = unique
    
    def validate(
        self,
        value: Any,
        path: str = "",
    ) -> ValidationResult:
        """Validate array."""
        null_result = self._check_null(value, path)
        if null_result:
            return null_result
        
        if not isinstance(value, list):
            return ValidationResult(
                is_valid=False,
                errors=[ValidationError(
                    f"Expected array, got {type(value).__name__}",
                    path=path,
                    value=value,
                )],
            )
        
        errors = []
        validated_items = []
        
        if self.min_items is not None and len(value) < self.min_items:
            errors.append(ValidationError(
                f"Array has {len(value)} items, minimum is {self.min_items}",
                path=path,
                value=value,
            ))
        
        if self.max_items is not None and len(value) > self.max_items:
            errors.append(ValidationError(
                f"Array has {len(value)} items, maximum is {self.max_items}",
                path=path,
                value=value,
            ))
        
        if self.unique and len(value) != len(set(str(v) for v in value)):
            errors.append(ValidationError(
                "Array items must be unique",
                path=path,
                value=value,
            ))
        
        # Validate each item
        if self.items:
            for i, item in enumerate(value):
                item_path = f"{path}[{i}]"
                result = self.items.validate(item, item_path)
                
                if not result.is_valid:
                    errors.extend(result.errors)
                else:
                    validated_items.append(result.data)
        else:
            validated_items = value
        
        return ValidationResult(
            is_valid=len(errors) == 0,
            data=validated_items,
            errors=errors,
        )


class Object(SchemaField):
    """Object schema field."""
    
    def __init__(
        self,
        properties: Optional[Dict[str, SchemaField]] = None,
        additional_properties: bool = True,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.properties = properties or {}
        self.additional_properties = additional_properties
    
    def validate(
        self,
        value: Any,
        path: str = "",
    ) -> ValidationResult:
        """Validate object."""
        null_result = self._check_null(value, path)
        if null_result:
            return null_result
        
        if not isinstance(value, dict):
            return ValidationResult(
                is_valid=False,
                errors=[ValidationError(
                    f"Expected object, got {type(value).__name__}",
                    path=path,
                    value=value,
                )],
            )
        
        errors = []
        validated_data = {}
        
        # Validate defined properties
        for prop_name, prop_schema in self.properties.items():
            prop_path = f"{path}.{prop_name}" if path else prop_name
            prop_value = value.get(prop_name)
            
            if prop_value is None and prop_name not in value:
                if prop_schema.required:
                    errors.append(ValidationError(
                        f"Missing required property: {prop_name}",
                        path=prop_path,
                    ))
                elif prop_schema.default is not None:
                    validated_data[prop_name] = prop_schema.default
            else:
                result = prop_schema.validate(prop_value, prop_path)
                
                if not result.is_valid:
                    errors.extend(result.errors)
                else:
                    validated_data[prop_name] = result.data
        
        # Handle additional properties
        for key in value:
            if key not in self.properties:
                if not self.additional_properties:
                    errors.append(ValidationError(
                        f"Additional property not allowed: {key}",
                        path=f"{path}.{key}" if path else key,
                    ))
                else:
                    validated_data[key] = value[key]
        
        return ValidationResult(
            is_valid=len(errors) == 0,
            data=validated_data,
            errors=errors,
        )


class Email(String):
    """Email schema field."""
    
    EMAIL_PATTERN = r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$'
    
    def __init__(self, **kwargs: Any):
        super().__init__(pattern=self.EMAIL_PATTERN, **kwargs)


class URL(String):
    """URL schema field."""
    
    URL_PATTERN = r'^https?://[^\s/$.?#].[^\s]*$'
    
    def __init__(self, **kwargs: Any):
        super().__init__(pattern=self.URL_PATTERN, **kwargs)


class UUID(String):
    """UUID schema field."""
    
    UUID_PATTERN = r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$'
    
    def __init__(self, **kwargs: Any):
        super().__init__(pattern=self.UUID_PATTERN, **kwargs)


class DateTime(SchemaField):
    """DateTime schema field."""
    
    def __init__(
        self,
        format: str = "%Y-%m-%dT%H:%M:%S",
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.format = format
    
    def validate(
        self,
        value: Any,
        path: str = "",
    ) -> ValidationResult:
        """Validate datetime."""
        null_result = self._check_null(value, path)
        if null_result:
            return null_result
        
        if isinstance(value, datetime):
            return ValidationResult(is_valid=True, data=value)
        
        if isinstance(value, str):
            try:
                parsed = datetime.strptime(value, self.format)
                return ValidationResult(is_valid=True, data=parsed)
            except ValueError:
                return ValidationResult(
                    is_valid=False,
                    errors=[ValidationError(
                        f"Invalid datetime format, expected: {self.format}",
                        path=path,
                        value=value,
                    )],
                )
        
        return ValidationResult(
            is_valid=False,
            errors=[ValidationError(
                f"Expected datetime, got {type(value).__name__}",
                path=path,
                value=value,
            )],
        )


class Enum_(SchemaField):
    """Enum schema field."""
    
    def __init__(
        self,
        values: List[Any],
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.values = values
    
    def validate(
        self,
        value: Any,
        path: str = "",
    ) -> ValidationResult:
        """Validate enum value."""
        null_result = self._check_null(value, path)
        if null_result:
            return null_result
        
        if value not in self.values:
            return ValidationResult(
                is_valid=False,
                errors=[ValidationError(
                    f"Value must be one of: {self.values}",
                    path=path,
                    value=value,
                )],
            )
        
        return ValidationResult(is_valid=True, data=value)


class Any_(SchemaField):
    """Any type schema field."""
    
    def validate(
        self,
        value: Any,
        path: str = "",
    ) -> ValidationResult:
        """Accept any value."""
        return ValidationResult(is_valid=True, data=value)


class Schema:
    """
    Main schema class for validating data structures.
    """
    
    def __init__(
        self,
        properties: Dict[str, SchemaField],
        additional_properties: bool = True,
        strict: bool = False,
    ):
        self._properties = properties
        self._additional_properties = additional_properties
        self._strict = strict
        self._object = Object(
            properties=properties,
            additional_properties=additional_properties,
        )
    
    def validate(
        self,
        data: Any,
    ) -> ValidationResult:
        """Validate data against schema."""
        return self._object.validate(data)
    
    def is_valid(self, data: Any) -> bool:
        """Check if data is valid."""
        return self.validate(data).is_valid
    
    def validate_or_raise(self, data: Any) -> Any:
        """Validate and raise on error."""
        result = self.validate(data)
        
        if not result.is_valid:
            error_messages = [e.message for e in result.errors]
            raise SchemaError(f"Validation failed: {'; '.join(error_messages)}")
        
        return result.data
    
    @classmethod
    def from_dict(cls, schema_dict: Dict[str, Any]) -> "Schema":
        """Create schema from dictionary."""
        properties = {}
        
        for name, field_def in schema_dict.items():
            if isinstance(field_def, SchemaField):
                properties[name] = field_def
            elif isinstance(field_def, dict):
                properties[name] = cls._field_from_dict(field_def)
            elif isinstance(field_def, type):
                properties[name] = cls._field_from_type(field_def)
        
        return cls(properties)
    
    @classmethod
    def _field_from_dict(cls, field_def: Dict[str, Any]) -> SchemaField:
        """Create field from dictionary definition."""
        field_type = field_def.get('type', 'any')
        
        type_map = {
            'string': String,
            'integer': Integer,
            'float': Float,
            'boolean': Boolean,
            'array': Array,
            'object': Object,
            'email': Email,
            'url': URL,
            'uuid': UUID,
            'datetime': DateTime,
        }
        
        field_cls = type_map.get(field_type, Any_)
        
        # Extract kwargs
        kwargs = {k: v for k, v in field_def.items() if k != 'type'}
        
        return field_cls(**kwargs)
    
    @classmethod
    def _field_from_type(cls, python_type: type) -> SchemaField:
        """Create field from Python type."""
        type_map = {
            str: String,
            int: Integer,
            float: Float,
            bool: Boolean,
            list: Array,
            dict: Object,
            datetime: DateTime,
        }
        
        return type_map.get(python_type, Any_)()


class SchemaBuilder:
    """
    Fluent schema builder.
    """
    
    def __init__(self):
        self._properties: Dict[str, SchemaField] = {}
        self._additional = True
    
    def string(
        self,
        name: str,
        **kwargs: Any,
    ) -> "SchemaBuilder":
        """Add string field."""
        self._properties[name] = String(**kwargs)
        return self
    
    def integer(
        self,
        name: str,
        **kwargs: Any,
    ) -> "SchemaBuilder":
        """Add integer field."""
        self._properties[name] = Integer(**kwargs)
        return self
    
    def float(
        self,
        name: str,
        **kwargs: Any,
    ) -> "SchemaBuilder":
        """Add float field."""
        self._properties[name] = Float(**kwargs)
        return self
    
    def boolean(
        self,
        name: str,
        **kwargs: Any,
    ) -> "SchemaBuilder":
        """Add boolean field."""
        self._properties[name] = Boolean(**kwargs)
        return self
    
    def array(
        self,
        name: str,
        items: Optional[SchemaField] = None,
        **kwargs: Any,
    ) -> "SchemaBuilder":
        """Add array field."""
        self._properties[name] = Array(items=items, **kwargs)
        return self
    
    def object(
        self,
        name: str,
        properties: Optional[Dict[str, SchemaField]] = None,
        **kwargs: Any,
    ) -> "SchemaBuilder":
        """Add object field."""
        self._properties[name] = Object(properties=properties, **kwargs)
        return self
    
    def email(
        self,
        name: str,
        **kwargs: Any,
    ) -> "SchemaBuilder":
        """Add email field."""
        self._properties[name] = Email(**kwargs)
        return self
    
    def field(
        self,
        name: str,
        schema_field: SchemaField,
    ) -> "SchemaBuilder":
        """Add custom field."""
        self._properties[name] = schema_field
        return self
    
    def no_additional_properties(self) -> "SchemaBuilder":
        """Disallow additional properties."""
        self._additional = False
        return self
    
    def build(self) -> Schema:
        """Build the schema."""
        return Schema(self._properties, self._additional)


# Decorators
def validate_schema(
    schema: Schema,
    on_error: Optional[Callable] = None,
) -> Callable:
    """
    Decorator to validate function arguments.
    
    Example:
        @validate_schema(UserSchema)
        def create_user(data: dict) -> User:
            return User(**data)
    """
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(data: Any, *args: Any, **kwargs: Any) -> Any:
            result = schema.validate(data)
            
            if not result.is_valid:
                if on_error:
                    return on_error(result.errors)
                raise SchemaError(
                    f"Validation failed: {result.errors[0].message}"
                )
            
            return func(result.data, *args, **kwargs)
        
        return wrapper
    
    return decorator


def dataclass_schema(cls: Type[T]) -> Type[T]:
    """
    Decorator to create schema from dataclass.
    
    Example:
        @dataclass_schema
        @dataclass
        class User:
            name: str
            age: int = 0
    """
    type_map = {
        str: String,
        int: Integer,
        float: Float,
        bool: Boolean,
        list: Array,
        dict: Object,
        datetime: DateTime,
    }
    
    properties = {}
    
    for f in dataclass_fields(cls):
        field_type = f.type
        
        # Handle Optional
        if hasattr(field_type, '__origin__'):
            if field_type.__origin__ is Union:
                # Get non-None type
                args = [a for a in field_type.__args__ if a is not type(None)]
                if args:
                    field_type = args[0]
        
        schema_field_cls = type_map.get(field_type, Any_)
        
        required = f.default is f.default_factory
        default = f.default if f.default is not f.default_factory else None
        
        properties[f.name] = schema_field_cls(
            required=required,
            default=default,
        )
    
    cls._schema = Schema(properties)
    
    original_init = cls.__init__
    
    @wraps(original_init)
    def validated_init(self, **kwargs):
        result = cls._schema.validate(kwargs)
        if not result.is_valid:
            raise SchemaError(result.errors[0].message)
        return original_init(self, **result.data)
    
    cls.__init__ = validated_init
    
    return cls


# Factory functions
def create_schema(
    properties: Dict[str, SchemaField],
    **kwargs: Any,
) -> Schema:
    """Create a schema."""
    return Schema(properties, **kwargs)


def create_schema_builder() -> SchemaBuilder:
    """Create a schema builder."""
    return SchemaBuilder()


def schema_from_dict(schema_dict: Dict[str, Any]) -> Schema:
    """Create schema from dictionary."""
    return Schema.from_dict(schema_dict)


__all__ = [
    # Exceptions
    "SchemaError",
    "ValidationError",
    # Enums
    "SchemaType",
    # Data classes
    "ValidationResult",
    "FieldInfo",
    # Schema fields
    "SchemaField",
    "String",
    "Integer",
    "Float",
    "Boolean",
    "Array",
    "Object",
    "Email",
    "URL",
    "UUID",
    "DateTime",
    "Enum_",
    "Any_",
    # Core classes
    "Schema",
    "SchemaBuilder",
    # Decorators
    "validate_schema",
    "dataclass_schema",
    # Factory functions
    "create_schema",
    "create_schema_builder",
    "schema_from_dict",
]



"""
Enterprise Approval System Module.

Multi-level approvals, routing, escalation,
delegation, and approval management.

Example:
    # Create approval system
    approvals = create_approval_system()
    
    # Define approval workflow
    workflow = await approvals.create_workflow(
        name="Expense Approval",
        levels=[
            {"role": "manager", "threshold": 1000},
            {"role": "director", "threshold": 5000},
            {"role": "cfo", "threshold": float("inf")},
        ],
    )
    
    # Submit for approval
    request = await approvals.submit(
        workflow_id=workflow.id,
        requester_id="user_123",
        title="Conference Travel",
        amount=2500,
        data={"purpose": "Tech conference"},
    )
    
    # Approve
    await approvals.approve(request.id, approver_id="manager_456", comments="Approved")
"""

from __future__ import annotations

import logging
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')

logger = logging.getLogger(__name__)


class ApprovalError(Exception):
    """Approval error."""
    pass


class RequestNotFoundError(ApprovalError):
    """Request not found."""
    pass


class ApprovalNotAllowedError(ApprovalError):
    """Approval not allowed."""
    pass


class ApprovalStatus(str, Enum):
    """Approval status."""
    PENDING = "pending"
    APPROVED = "approved"
    REJECTED = "rejected"
    ESCALATED = "escalated"
    DELEGATED = "delegated"
    CANCELLED = "cancelled"
    EXPIRED = "expired"


class RequestStatus(str, Enum):
    """Request status."""
    DRAFT = "draft"
    SUBMITTED = "submitted"
    IN_PROGRESS = "in_progress"
    APPROVED = "approved"
    REJECTED = "rejected"
    CANCELLED = "cancelled"


class EscalationReason(str, Enum):
    """Escalation reason."""
    TIMEOUT = "timeout"
    MANUAL = "manual"
    THRESHOLD = "threshold"
    POLICY = "policy"


@dataclass
class ApprovalLevel:
    """Approval level."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    order: int = 0
    name: str = ""
    role: str = ""
    approvers: List[str] = field(default_factory=list)
    required_count: int = 1  # How many approvers needed
    threshold: float = float("inf")  # Amount threshold
    timeout_hours: int = 72
    auto_approve: bool = False
    auto_escalate: bool = True
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class ApprovalWorkflow:
    """Approval workflow."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    description: str = ""
    levels: List[ApprovalLevel] = field(default_factory=list)
    require_sequential: bool = True
    allow_delegation: bool = True
    allow_recall: bool = True
    notify_on_action: bool = True
    created_by: str = ""
    created_at: datetime = field(default_factory=datetime.utcnow)
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class ApprovalAction:
    """Approval action."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    level_id: str = ""
    approver_id: str = ""
    status: ApprovalStatus = ApprovalStatus.PENDING
    comments: str = ""
    delegated_to: Optional[str] = None
    delegated_by: Optional[str] = None
    created_at: datetime = field(default_factory=datetime.utcnow)
    completed_at: Optional[datetime] = None
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class ApprovalRequest:
    """Approval request."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    workflow_id: str = ""
    workflow_name: str = ""
    status: RequestStatus = RequestStatus.DRAFT
    requester_id: str = ""
    title: str = ""
    description: str = ""
    amount: float = 0.0
    data: Dict[str, Any] = field(default_factory=dict)
    current_level: int = 0
    actions: List[ApprovalAction] = field(default_factory=list)
    attachments: List[str] = field(default_factory=list)
    due_date: Optional[datetime] = None
    submitted_at: Optional[datetime] = None
    completed_at: Optional[datetime] = None
    final_status: Optional[ApprovalStatus] = None
    final_comments: str = ""
    correlation_id: str = ""
    metadata: Dict[str, Any] = field(default_factory=dict)
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class ApprovalStats:
    """Approval statistics."""
    total_requests: int = 0
    pending_requests: int = 0
    approved_requests: int = 0
    rejected_requests: int = 0
    average_approval_time_hours: float = 0.0


# Workflow store
class WorkflowStore(ABC):
    """Workflow storage."""
    
    @abstractmethod
    async def save(self, workflow: ApprovalWorkflow) -> None:
        pass
    
    @abstractmethod
    async def get(self, workflow_id: str) -> Optional[ApprovalWorkflow]:
        pass
    
    @abstractmethod
    async def list(self) -> List[ApprovalWorkflow]:
        pass


class InMemoryWorkflowStore(WorkflowStore):
    """In-memory workflow store."""
    
    def __init__(self):
        self._workflows: Dict[str, ApprovalWorkflow] = {}
    
    async def save(self, workflow: ApprovalWorkflow) -> None:
        self._workflows[workflow.id] = workflow
    
    async def get(self, workflow_id: str) -> Optional[ApprovalWorkflow]:
        return self._workflows.get(workflow_id)
    
    async def list(self) -> List[ApprovalWorkflow]:
        return list(self._workflows.values())


# Request store
class RequestStore(ABC):
    """Request storage."""
    
    @abstractmethod
    async def save(self, request: ApprovalRequest) -> None:
        pass
    
    @abstractmethod
    async def get(self, request_id: str) -> Optional[ApprovalRequest]:
        pass
    
    @abstractmethod
    async def query(
        self,
        status: Optional[RequestStatus] = None,
        requester_id: Optional[str] = None,
        approver_id: Optional[str] = None,
    ) -> List[ApprovalRequest]:
        pass


class InMemoryRequestStore(RequestStore):
    """In-memory request store."""
    
    def __init__(self):
        self._requests: Dict[str, ApprovalRequest] = {}
    
    async def save(self, request: ApprovalRequest) -> None:
        self._requests[request.id] = request
    
    async def get(self, request_id: str) -> Optional[ApprovalRequest]:
        return self._requests.get(request_id)
    
    async def query(
        self,
        status: Optional[RequestStatus] = None,
        requester_id: Optional[str] = None,
        approver_id: Optional[str] = None,
    ) -> List[ApprovalRequest]:
        results = list(self._requests.values())
        
        if status:
            results = [r for r in results if r.status == status]
        
        if requester_id:
            results = [r for r in results if r.requester_id == requester_id]
        
        if approver_id:
            results = [
                r for r in results
                if any(a.approver_id == approver_id for a in r.actions)
            ]
        
        return sorted(results, key=lambda r: r.created_at, reverse=True)


# Approval system
class ApprovalSystem:
    """Approval system."""
    
    def __init__(
        self,
        workflow_store: Optional[WorkflowStore] = None,
        request_store: Optional[RequestStore] = None,
    ):
        self._workflows = workflow_store or InMemoryWorkflowStore()
        self._requests = request_store or InMemoryRequestStore()
        self._stats = ApprovalStats()
        self._approval_times: List[float] = []
    
    async def create_workflow(
        self,
        name: str,
        levels: List[Dict[str, Any]] = None,
        description: str = "",
        created_by: str = "",
        **kwargs,
    ) -> ApprovalWorkflow:
        """Create approval workflow."""
        workflow = ApprovalWorkflow(
            name=name,
            description=description,
            created_by=created_by,
            **kwargs,
        )
        
        # Add levels
        if levels:
            for i, level_def in enumerate(levels):
                level = ApprovalLevel(
                    order=i,
                    name=level_def.get("name", f"Level {i + 1}"),
                    role=level_def.get("role", ""),
                    approvers=level_def.get("approvers", []),
                    required_count=level_def.get("required_count", 1),
                    threshold=level_def.get("threshold", float("inf")),
                    timeout_hours=level_def.get("timeout_hours", 72),
                )
                workflow.levels.append(level)
        
        await self._workflows.save(workflow)
        
        logger.info(f"Approval workflow created: {name}")
        
        return workflow
    
    async def get_workflow(self, workflow_id: str) -> Optional[ApprovalWorkflow]:
        """Get workflow."""
        return await self._workflows.get(workflow_id)
    
    async def submit(
        self,
        workflow_id: str,
        requester_id: str,
        title: str,
        amount: float = 0.0,
        description: str = "",
        data: Dict[str, Any] = None,
        **kwargs,
    ) -> ApprovalRequest:
        """Submit approval request."""
        workflow = await self._workflows.get(workflow_id)
        if not workflow:
            raise ApprovalError(f"Workflow not found: {workflow_id}")
        
        request = ApprovalRequest(
            workflow_id=workflow_id,
            workflow_name=workflow.name,
            requester_id=requester_id,
            title=title,
            description=description,
            amount=amount,
            data=data or {},
            status=RequestStatus.SUBMITTED,
            submitted_at=datetime.utcnow(),
            **kwargs,
        )
        
        # Determine starting level based on amount
        request.current_level = self._get_starting_level(workflow, amount)
        
        # Create pending actions for current level
        current_level = workflow.levels[request.current_level]
        self._create_level_actions(request, current_level)
        
        await self._requests.save(request)
        
        self._stats.total_requests += 1
        self._stats.pending_requests += 1
        
        logger.info(f"Approval request submitted: {title}")
        
        return request
    
    def _get_starting_level(
        self,
        workflow: ApprovalWorkflow,
        amount: float,
    ) -> int:
        """Get starting approval level based on amount."""
        for i, level in enumerate(workflow.levels):
            if amount <= level.threshold:
                return i
        return len(workflow.levels) - 1
    
    def _create_level_actions(
        self,
        request: ApprovalRequest,
        level: ApprovalLevel,
    ) -> None:
        """Create pending actions for level."""
        for approver_id in level.approvers:
            request.actions.append(ApprovalAction(
                level_id=level.id,
                approver_id=approver_id,
                status=ApprovalStatus.PENDING,
            ))
    
    async def get_request(
        self,
        request_id: str,
    ) -> Optional[ApprovalRequest]:
        """Get request."""
        return await self._requests.get(request_id)
    
    async def approve(
        self,
        request_id: str,
        approver_id: str,
        comments: str = "",
    ) -> ApprovalRequest:
        """Approve request."""
        request = await self._requests.get(request_id)
        if not request:
            raise RequestNotFoundError(f"Request not found: {request_id}")
        
        if request.status not in (RequestStatus.SUBMITTED, RequestStatus.IN_PROGRESS):
            raise ApprovalNotAllowedError(f"Request cannot be approved: {request.status}")
        
        workflow = await self._workflows.get(request.workflow_id)
        if not workflow:
            raise ApprovalError("Workflow not found")
        
        # Find and update action
        action = self._find_pending_action(request, approver_id)
        if not action:
            raise ApprovalNotAllowedError("No pending approval for this user")
        
        action.status = ApprovalStatus.APPROVED
        action.comments = comments
        action.completed_at = datetime.utcnow()
        
        # Check if level is complete
        current_level = workflow.levels[request.current_level]
        level_approved = self._is_level_approved(request, current_level)
        
        if level_approved:
            # Move to next level or complete
            if request.current_level < len(workflow.levels) - 1:
                request.current_level += 1
                request.status = RequestStatus.IN_PROGRESS
                next_level = workflow.levels[request.current_level]
                self._create_level_actions(request, next_level)
            else:
                request.status = RequestStatus.APPROVED
                request.completed_at = datetime.utcnow()
                request.final_status = ApprovalStatus.APPROVED
                request.final_comments = comments
                
                self._stats.pending_requests -= 1
                self._stats.approved_requests += 1
                
                # Track approval time
                if request.submitted_at:
                    hours = (request.completed_at - request.submitted_at).total_seconds() / 3600
                    self._approval_times.append(hours)
                    self._stats.average_approval_time_hours = sum(self._approval_times) / len(self._approval_times)
        
        await self._requests.save(request)
        
        logger.info(f"Request approved by {approver_id}: {request.title}")
        
        return request
    
    def _find_pending_action(
        self,
        request: ApprovalRequest,
        approver_id: str,
    ) -> Optional[ApprovalAction]:
        """Find pending action for approver."""
        for action in request.actions:
            if action.approver_id == approver_id and action.status == ApprovalStatus.PENDING:
                return action
        return None
    
    def _is_level_approved(
        self,
        request: ApprovalRequest,
        level: ApprovalLevel,
    ) -> bool:
        """Check if level has enough approvals."""
        approved_count = sum(
            1 for a in request.actions
            if a.level_id == level.id and a.status == ApprovalStatus.APPROVED
        )
        return approved_count >= level.required_count
    
    async def reject(
        self,
        request_id: str,
        approver_id: str,
        comments: str = "",
    ) -> ApprovalRequest:
        """Reject request."""
        request = await self._requests.get(request_id)
        if not request:
            raise RequestNotFoundError(f"Request not found: {request_id}")
        
        if request.status not in (RequestStatus.SUBMITTED, RequestStatus.IN_PROGRESS):
            raise ApprovalNotAllowedError(f"Request cannot be rejected: {request.status}")
        
        action = self._find_pending_action(request, approver_id)
        if not action:
            raise ApprovalNotAllowedError("No pending approval for this user")
        
        action.status = ApprovalStatus.REJECTED
        action.comments = comments
        action.completed_at = datetime.utcnow()
        
        request.status = RequestStatus.REJECTED
        request.completed_at = datetime.utcnow()
        request.final_status = ApprovalStatus.REJECTED
        request.final_comments = comments
        
        await self._requests.save(request)
        
        self._stats.pending_requests -= 1
        self._stats.rejected_requests += 1
        
        logger.info(f"Request rejected by {approver_id}: {request.title}")
        
        return request
    
    async def delegate(
        self,
        request_id: str,
        approver_id: str,
        delegate_to: str,
        comments: str = "",
    ) -> ApprovalRequest:
        """Delegate approval."""
        request = await self._requests.get(request_id)
        if not request:
            raise RequestNotFoundError(f"Request not found: {request_id}")
        
        workflow = await self._workflows.get(request.workflow_id)
        if not workflow or not workflow.allow_delegation:
            raise ApprovalNotAllowedError("Delegation not allowed")
        
        action = self._find_pending_action(request, approver_id)
        if not action:
            raise ApprovalNotAllowedError("No pending approval for this user")
        
        action.status = ApprovalStatus.DELEGATED
        action.delegated_to = delegate_to
        action.delegated_by = approver_id
        action.comments = comments
        action.completed_at = datetime.utcnow()
        
        # Create new action for delegate
        request.actions.append(ApprovalAction(
            level_id=action.level_id,
            approver_id=delegate_to,
            status=ApprovalStatus.PENDING,
        ))
        
        await self._requests.save(request)
        
        logger.info(f"Request delegated from {approver_id} to {delegate_to}")
        
        return request
    
    async def escalate(
        self,
        request_id: str,
        reason: EscalationReason = EscalationReason.MANUAL,
        comments: str = "",
    ) -> ApprovalRequest:
        """Escalate request to next level."""
        request = await self._requests.get(request_id)
        if not request:
            raise RequestNotFoundError(f"Request not found: {request_id}")
        
        workflow = await self._workflows.get(request.workflow_id)
        if not workflow:
            raise ApprovalError("Workflow not found")
        
        if request.current_level >= len(workflow.levels) - 1:
            raise ApprovalNotAllowedError("Already at highest level")
        
        # Mark current actions as escalated
        current_level = workflow.levels[request.current_level]
        for action in request.actions:
            if action.level_id == current_level.id and action.status == ApprovalStatus.PENDING:
                action.status = ApprovalStatus.ESCALATED
                action.comments = f"Escalated: {reason.value}"
                action.completed_at = datetime.utcnow()
        
        # Move to next level
        request.current_level += 1
        request.status = RequestStatus.IN_PROGRESS
        
        next_level = workflow.levels[request.current_level]
        self._create_level_actions(request, next_level)
        
        await self._requests.save(request)
        
        logger.info(f"Request escalated: {request.title}")
        
        return request
    
    async def cancel(
        self,
        request_id: str,
        comments: str = "",
    ) -> ApprovalRequest:
        """Cancel request."""
        request = await self._requests.get(request_id)
        if not request:
            raise RequestNotFoundError(f"Request not found: {request_id}")
        
        if request.status in (RequestStatus.APPROVED, RequestStatus.REJECTED):
            raise ApprovalNotAllowedError("Cannot cancel completed request")
        
        request.status = RequestStatus.CANCELLED
        request.completed_at = datetime.utcnow()
        request.final_status = ApprovalStatus.CANCELLED
        request.final_comments = comments
        
        await self._requests.save(request)
        
        self._stats.pending_requests -= 1
        
        logger.info(f"Request cancelled: {request.title}")
        
        return request
    
    async def get_pending_approvals(
        self,
        approver_id: str,
    ) -> List[ApprovalRequest]:
        """Get pending approvals for user."""
        requests = await self._requests.query(status=RequestStatus.SUBMITTED)
        requests.extend(await self._requests.query(status=RequestStatus.IN_PROGRESS))
        
        pending = []
        for request in requests:
            if self._find_pending_action(request, approver_id):
                pending.append(request)
        
        return pending
    
    async def query_requests(
        self,
        status: Optional[RequestStatus] = None,
        requester_id: Optional[str] = None,
    ) -> List[ApprovalRequest]:
        """Query requests."""
        return await self._requests.query(status=status, requester_id=requester_id)
    
    def get_stats(self) -> ApprovalStats:
        """Get statistics."""
        return self._stats


# Factory functions
def create_approval_system() -> ApprovalSystem:
    """Create approval system."""
    return ApprovalSystem()


def create_workflow(
    name: str,
    **kwargs,
) -> ApprovalWorkflow:
    """Create workflow."""
    return ApprovalWorkflow(name=name, **kwargs)


def create_level(
    name: str,
    role: str,
    **kwargs,
) -> ApprovalLevel:
    """Create approval level."""
    return ApprovalLevel(name=name, role=role, **kwargs)


__all__ = [
    # Exceptions
    "ApprovalError",
    "RequestNotFoundError",
    "ApprovalNotAllowedError",
    # Enums
    "ApprovalStatus",
    "RequestStatus",
    "EscalationReason",
    # Data classes
    "ApprovalLevel",
    "ApprovalWorkflow",
    "ApprovalAction",
    "ApprovalRequest",
    "ApprovalStats",
    # Stores
    "WorkflowStore",
    "InMemoryWorkflowStore",
    "RequestStore",
    "InMemoryRequestStore",
    # System
    "ApprovalSystem",
    # Factory functions
    "create_approval_system",
    "create_workflow",
    "create_level",
]



"""
Enterprise Job Queue Manager Module.

Async job execution, worker pools, priority queues,
job scheduling, and retry management.

Example:
    # Create job queue manager
    jobs = create_job_queue_manager()
    
    # Define job handler
    @jobs.handler("send_email")
    async def send_email(job: Job):
        await email_service.send(job.payload["to"], job.payload["subject"])
    
    # Submit job
    job = await jobs.submit(
        queue="send_email",
        payload={"to": "user@example.com", "subject": "Hello"},
    )
    
    # Start workers
    await jobs.start_workers(queue="send_email", count=5)
"""

from __future__ import annotations

import asyncio
import logging
import random
import time
import traceback
import uuid
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from heapq import heappop, heappush
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    TypeVar,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")


class JobError(Exception):
    """Job error."""
    pass


class JobNotFoundError(JobError):
    """Job not found error."""
    pass


class QueueNotFoundError(JobError):
    """Queue not found error."""
    pass


class JobStatus(str, Enum):
    """Job status."""
    PENDING = "pending"
    SCHEDULED = "scheduled"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    RETRYING = "retrying"
    CANCELLED = "cancelled"
    DEAD = "dead"


class JobPriority(int, Enum):
    """Job priority."""
    LOW = 10
    NORMAL = 50
    HIGH = 80
    CRITICAL = 100


class RetryStrategy(str, Enum):
    """Retry strategy."""
    FIXED = "fixed"
    LINEAR = "linear"
    EXPONENTIAL = "exponential"
    CUSTOM = "custom"


@dataclass
class RetryPolicy:
    """Retry policy."""
    max_retries: int = 3
    strategy: RetryStrategy = RetryStrategy.EXPONENTIAL
    base_delay_seconds: float = 1.0
    max_delay_seconds: float = 300.0
    retry_on_exceptions: List[str] = field(default_factory=list)


@dataclass
class Job:
    """Job."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    queue: str = ""
    
    # Payload
    payload: Dict[str, Any] = field(default_factory=dict)
    
    # Status
    status: JobStatus = JobStatus.PENDING
    
    # Priority and scheduling
    priority: int = JobPriority.NORMAL
    scheduled_at: Optional[datetime] = None
    delay_seconds: float = 0.0
    
    # Progress
    progress: float = 0.0
    progress_message: str = ""
    
    # Result
    result: Any = None
    error_message: str = ""
    error_traceback: str = ""
    
    # Retry
    attempt: int = 0
    max_retries: int = 3
    next_retry_at: Optional[datetime] = None
    
    # Timing
    created_at: datetime = field(default_factory=datetime.utcnow)
    started_at: Optional[datetime] = None
    completed_at: Optional[datetime] = None
    duration_seconds: float = 0.0
    
    # Metadata
    tags: List[str] = field(default_factory=list)
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    def __lt__(self, other: "Job") -> bool:
        """Compare for priority queue."""
        return self.priority > other.priority  # Higher priority first


@dataclass
class QueueConfig:
    """Queue configuration."""
    name: str = ""
    
    # Workers
    max_workers: int = 5
    min_workers: int = 1
    
    # Concurrency
    max_concurrent_jobs: int = 10
    
    # Timeouts
    job_timeout_seconds: float = 300.0
    poll_interval_seconds: float = 1.0
    
    # Retry
    retry_policy: RetryPolicy = field(default_factory=RetryPolicy)
    
    # Dead letter
    dead_letter_queue: Optional[str] = None
    
    # Metadata
    description: str = ""


@dataclass
class Worker:
    """Worker."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    queue: str = ""
    
    # Status
    status: str = "idle"  # idle, busy, stopped
    current_job_id: Optional[str] = None
    
    # Statistics
    jobs_processed: int = 0
    jobs_failed: int = 0
    
    # Timing
    started_at: datetime = field(default_factory=datetime.utcnow)
    last_job_at: Optional[datetime] = None
    
    # Task
    task: Optional[asyncio.Task] = None


@dataclass
class QueueStats:
    """Queue statistics."""
    queue_name: str = ""
    
    pending_jobs: int = 0
    running_jobs: int = 0
    completed_jobs: int = 0
    failed_jobs: int = 0
    
    workers_count: int = 0
    active_workers: int = 0
    
    avg_wait_time_seconds: float = 0.0
    avg_processing_time_seconds: float = 0.0
    throughput_per_minute: float = 0.0


@dataclass
class JobQueueStats:
    """Overall job queue statistics."""
    total_queues: int = 0
    total_jobs: int = 0
    pending_jobs: int = 0
    running_jobs: int = 0
    total_workers: int = 0


# Job store
class JobStore(ABC):
    """Job storage."""
    
    @abstractmethod
    async def save(self, job: Job) -> None:
        pass
    
    @abstractmethod
    async def get(self, job_id: str) -> Optional[Job]:
        pass
    
    @abstractmethod
    async def get_pending(self, queue: str, limit: int) -> List[Job]:
        pass
    
    @abstractmethod
    async def get_scheduled(self, before: datetime) -> List[Job]:
        pass
    
    @abstractmethod
    async def count_by_status(self, queue: str, status: JobStatus) -> int:
        pass


class InMemoryJobStore(JobStore):
    """In-memory job store."""
    
    def __init__(self, max_jobs: int = 100000):
        self._jobs: Dict[str, Job] = {}
        self._queues: Dict[str, List[str]] = defaultdict(list)
        self._max_jobs = max_jobs
    
    async def save(self, job: Job) -> None:
        self._jobs[job.id] = job
        
        if job.id not in self._queues[job.queue]:
            self._queues[job.queue].append(job.id)
        
        # Cleanup old completed jobs
        if len(self._jobs) > self._max_jobs:
            self._cleanup_old_jobs()
    
    async def get(self, job_id: str) -> Optional[Job]:
        return self._jobs.get(job_id)
    
    async def get_pending(self, queue: str, limit: int) -> List[Job]:
        job_ids = self._queues.get(queue, [])
        jobs = []
        
        for jid in job_ids:
            job = self._jobs.get(jid)
            if job and job.status == JobStatus.PENDING:
                jobs.append(job)
                if len(jobs) >= limit:
                    break
        
        return sorted(jobs, key=lambda j: j.priority, reverse=True)
    
    async def get_scheduled(self, before: datetime) -> List[Job]:
        jobs = []
        
        for job in self._jobs.values():
            if job.status == JobStatus.SCHEDULED:
                if job.scheduled_at and job.scheduled_at <= before:
                    jobs.append(job)
        
        return jobs
    
    async def count_by_status(self, queue: str, status: JobStatus) -> int:
        job_ids = self._queues.get(queue, [])
        count = 0
        
        for jid in job_ids:
            job = self._jobs.get(jid)
            if job and job.status == status:
                count += 1
        
        return count
    
    def _cleanup_old_jobs(self) -> None:
        completed = [
            j for j in self._jobs.values()
            if j.status in (JobStatus.COMPLETED, JobStatus.DEAD, JobStatus.CANCELLED)
        ]
        
        completed.sort(key=lambda j: j.completed_at or j.created_at)
        
        to_remove = len(self._jobs) - (self._max_jobs // 2)
        
        for job in completed[:to_remove]:
            self._jobs.pop(job.id, None)
            if job.id in self._queues.get(job.queue, []):
                self._queues[job.queue].remove(job.id)


# Job handler
JobHandler = Callable[[Job], Any]


# Job queue
class JobQueue:
    """Job queue."""
    
    def __init__(
        self,
        config: QueueConfig,
        store: JobStore,
    ):
        self._config = config
        self._store = store
        
        self._handlers: Dict[str, JobHandler] = {}
        self._workers: Dict[str, Worker] = {}
        
        self._running = False
        self._job_semaphore = asyncio.Semaphore(config.max_concurrent_jobs)
        
        # Priority queue for pending jobs
        self._priority_queue: List[Job] = []
        self._queue_lock = asyncio.Lock()
        
        # Statistics
        self._completed_count = 0
        self._failed_count = 0
        self._processing_times: List[float] = []
        self._wait_times: List[float] = []
    
    @property
    def name(self) -> str:
        return self._config.name
    
    async def submit(
        self,
        payload: Dict[str, Any],
        priority: int = JobPriority.NORMAL,
        delay_seconds: float = 0.0,
        scheduled_at: Optional[datetime] = None,
        **kwargs,
    ) -> Job:
        """Submit a job."""
        job = Job(
            queue=self._config.name,
            payload=payload,
            priority=priority,
            delay_seconds=delay_seconds,
            scheduled_at=scheduled_at,
            max_retries=self._config.retry_policy.max_retries,
            **kwargs,
        )
        
        if scheduled_at:
            job.status = JobStatus.SCHEDULED
        elif delay_seconds > 0:
            job.scheduled_at = datetime.utcnow() + timedelta(seconds=delay_seconds)
            job.status = JobStatus.SCHEDULED
        
        await self._store.save(job)
        
        if job.status == JobStatus.PENDING:
            async with self._queue_lock:
                heappush(self._priority_queue, job)
        
        return job
    
    async def get_job(self, job_id: str) -> Optional[Job]:
        """Get job by ID."""
        return await self._store.get(job_id)
    
    async def cancel_job(self, job_id: str) -> bool:
        """Cancel a job."""
        job = await self._store.get(job_id)
        
        if not job:
            return False
        
        if job.status in (JobStatus.PENDING, JobStatus.SCHEDULED, JobStatus.RETRYING):
            job.status = JobStatus.CANCELLED
            job.completed_at = datetime.utcnow()
            await self._store.save(job)
            return True
        
        return False
    
    def register_handler(self, handler: JobHandler) -> None:
        """Register job handler."""
        self._handlers[self._config.name] = handler
    
    async def start_workers(self, count: Optional[int] = None) -> List[Worker]:
        """Start worker tasks."""
        count = count or self._config.min_workers
        count = min(count, self._config.max_workers)
        
        self._running = True
        workers = []
        
        for _ in range(count):
            worker = Worker(queue=self._config.name)
            worker.task = asyncio.create_task(self._worker_loop(worker))
            self._workers[worker.id] = worker
            workers.append(worker)
        
        logger.info(f"Started {count} workers for queue: {self._config.name}")
        
        return workers
    
    async def stop_workers(self) -> None:
        """Stop all workers."""
        self._running = False
        
        for worker in self._workers.values():
            if worker.task:
                worker.task.cancel()
                try:
                    await worker.task
                except asyncio.CancelledError:
                    pass
            worker.status = "stopped"
        
        logger.info(f"Stopped workers for queue: {self._config.name}")
    
    async def scale_workers(self, count: int) -> None:
        """Scale workers to specified count."""
        current = len(self._workers)
        
        if count > current:
            # Add workers
            for _ in range(count - current):
                worker = Worker(queue=self._config.name)
                worker.task = asyncio.create_task(self._worker_loop(worker))
                self._workers[worker.id] = worker
        
        elif count < current:
            # Remove workers
            workers_to_remove = list(self._workers.values())[count:]
            for worker in workers_to_remove:
                if worker.task:
                    worker.task.cancel()
                self._workers.pop(worker.id, None)
    
    async def get_stats(self) -> QueueStats:
        """Get queue statistics."""
        pending = await self._store.count_by_status(self._config.name, JobStatus.PENDING)
        running = await self._store.count_by_status(self._config.name, JobStatus.RUNNING)
        
        active_workers = len([w for w in self._workers.values() if w.status == "busy"])
        
        avg_wait = sum(self._wait_times) / len(self._wait_times) if self._wait_times else 0
        avg_proc = sum(self._processing_times) / len(self._processing_times) if self._processing_times else 0
        
        return QueueStats(
            queue_name=self._config.name,
            pending_jobs=pending,
            running_jobs=running,
            completed_jobs=self._completed_count,
            failed_jobs=self._failed_count,
            workers_count=len(self._workers),
            active_workers=active_workers,
            avg_wait_time_seconds=avg_wait,
            avg_processing_time_seconds=avg_proc,
        )
    
    async def _worker_loop(self, worker: Worker) -> None:
        """Worker loop."""
        while self._running:
            try:
                job = await self._get_next_job()
                
                if not job:
                    await asyncio.sleep(self._config.poll_interval_seconds)
                    continue
                
                worker.status = "busy"
                worker.current_job_id = job.id
                
                await self._process_job(job, worker)
                
                worker.status = "idle"
                worker.current_job_id = None
                worker.last_job_at = datetime.utcnow()
                
            except asyncio.CancelledError:
                break
            except Exception as e:
                logger.error(f"Worker error: {e}")
                await asyncio.sleep(1)
    
    async def _get_next_job(self) -> Optional[Job]:
        """Get next job to process."""
        # Check scheduled jobs
        scheduled = await self._store.get_scheduled(datetime.utcnow())
        for job in scheduled:
            job.status = JobStatus.PENDING
            await self._store.save(job)
            async with self._queue_lock:
                heappush(self._priority_queue, job)
        
        # Get from priority queue
        async with self._queue_lock:
            if self._priority_queue:
                return heappop(self._priority_queue)
        
        # Fallback to store
        pending = await self._store.get_pending(self._config.name, 1)
        return pending[0] if pending else None
    
    async def _process_job(self, job: Job, worker: Worker) -> None:
        """Process a job."""
        handler = self._handlers.get(self._config.name)
        
        if not handler:
            logger.error(f"No handler for queue: {self._config.name}")
            return
        
        job.status = JobStatus.RUNNING
        job.started_at = datetime.utcnow()
        job.attempt += 1
        await self._store.save(job)
        
        wait_time = (job.started_at - job.created_at).total_seconds()
        self._wait_times.append(wait_time)
        if len(self._wait_times) > 1000:
            self._wait_times = self._wait_times[-1000:]
        
        start_time = time.monotonic()
        
        try:
            async with self._job_semaphore:
                if asyncio.iscoroutinefunction(handler):
                    result = await asyncio.wait_for(
                        handler(job),
                        timeout=self._config.job_timeout_seconds,
                    )
                else:
                    result = handler(job)
            
            job.status = JobStatus.COMPLETED
            job.result = result
            job.completed_at = datetime.utcnow()
            
            worker.jobs_processed += 1
            self._completed_count += 1
            
        except asyncio.TimeoutError:
            await self._handle_failure(
                job, worker, f"Job timed out after {self._config.job_timeout_seconds}s"
            )
        
        except Exception as e:
            await self._handle_failure(job, worker, str(e), traceback.format_exc())
        
        finally:
            job.duration_seconds = time.monotonic() - start_time
            self._processing_times.append(job.duration_seconds)
            if len(self._processing_times) > 1000:
                self._processing_times = self._processing_times[-1000:]
            
            await self._store.save(job)
    
    async def _handle_failure(
        self,
        job: Job,
        worker: Worker,
        error_message: str,
        error_traceback: str = "",
    ) -> None:
        """Handle job failure."""
        job.error_message = error_message
        job.error_traceback = error_traceback
        
        worker.jobs_failed += 1
        
        if job.attempt < job.max_retries:
            # Schedule retry
            delay = self._calculate_retry_delay(job.attempt)
            job.status = JobStatus.RETRYING
            job.next_retry_at = datetime.utcnow() + timedelta(seconds=delay)
            job.scheduled_at = job.next_retry_at
            
            logger.warning(
                f"Job {job.id} failed, retrying in {delay}s (attempt {job.attempt})"
            )
        else:
            # Move to dead letter or mark as dead
            job.status = JobStatus.DEAD
            job.completed_at = datetime.utcnow()
            
            self._failed_count += 1
            
            logger.error(f"Job {job.id} failed permanently after {job.attempt} attempts")
    
    def _calculate_retry_delay(self, attempt: int) -> float:
        """Calculate retry delay."""
        policy = self._config.retry_policy
        
        if policy.strategy == RetryStrategy.FIXED:
            delay = policy.base_delay_seconds
        
        elif policy.strategy == RetryStrategy.LINEAR:
            delay = policy.base_delay_seconds * attempt
        
        elif policy.strategy == RetryStrategy.EXPONENTIAL:
            delay = policy.base_delay_seconds * (2 ** (attempt - 1))
        
        else:
            delay = policy.base_delay_seconds
        
        # Add jitter
        jitter = random.uniform(0, delay * 0.1)
        delay = delay + jitter
        
        return min(delay, policy.max_delay_seconds)


# Job queue manager
class JobQueueManager:
    """Job queue manager."""
    
    def __init__(
        self,
        store: Optional[JobStore] = None,
    ):
        self._store = store or InMemoryJobStore()
        self._queues: Dict[str, JobQueue] = {}
        self._handlers: Dict[str, JobHandler] = {}
        self._listeners: List[Callable] = []
    
    async def create_queue(
        self,
        name: str,
        max_workers: int = 5,
        max_concurrent_jobs: int = 10,
        job_timeout_seconds: float = 300.0,
        retry_policy: Optional[RetryPolicy] = None,
        **kwargs,
    ) -> JobQueue:
        """Create a job queue."""
        config = QueueConfig(
            name=name,
            max_workers=max_workers,
            max_concurrent_jobs=max_concurrent_jobs,
            job_timeout_seconds=job_timeout_seconds,
            retry_policy=retry_policy or RetryPolicy(),
            **kwargs,
        )
        
        queue = JobQueue(config, self._store)
        self._queues[name] = queue
        
        # Register any pre-registered handlers
        if name in self._handlers:
            queue.register_handler(self._handlers[name])
        
        logger.info(f"Queue created: {name}")
        
        return queue
    
    async def get_queue(self, name: str) -> Optional[JobQueue]:
        """Get queue by name."""
        return self._queues.get(name)
    
    async def list_queues(self) -> List[str]:
        """List queue names."""
        return list(self._queues.keys())
    
    async def delete_queue(self, name: str) -> bool:
        """Delete a queue."""
        queue = self._queues.pop(name, None)
        
        if queue:
            await queue.stop_workers()
            return True
        
        return False
    
    def handler(self, queue_name: str) -> Callable:
        """Decorator to register job handler."""
        def decorator(func: JobHandler) -> JobHandler:
            self._handlers[queue_name] = func
            
            queue = self._queues.get(queue_name)
            if queue:
                queue.register_handler(func)
            
            return func
        
        return decorator
    
    async def submit(
        self,
        queue: str,
        payload: Dict[str, Any],
        priority: int = JobPriority.NORMAL,
        delay_seconds: float = 0.0,
        scheduled_at: Optional[datetime] = None,
        **kwargs,
    ) -> Job:
        """Submit a job to a queue."""
        job_queue = self._queues.get(queue)
        
        if not job_queue:
            # Auto-create queue
            job_queue = await self.create_queue(queue)
        
        return await job_queue.submit(
            payload=payload,
            priority=priority,
            delay_seconds=delay_seconds,
            scheduled_at=scheduled_at,
            **kwargs,
        )
    
    async def get_job(self, job_id: str) -> Optional[Job]:
        """Get job by ID."""
        return await self._store.get(job_id)
    
    async def cancel_job(self, job_id: str) -> bool:
        """Cancel a job."""
        job = await self._store.get(job_id)
        
        if not job:
            return False
        
        queue = self._queues.get(job.queue)
        
        if queue:
            return await queue.cancel_job(job_id)
        
        return False
    
    async def start_workers(
        self,
        queue: str,
        count: Optional[int] = None,
    ) -> List[Worker]:
        """Start workers for a queue."""
        job_queue = self._queues.get(queue)
        
        if not job_queue:
            raise QueueNotFoundError(f"Queue not found: {queue}")
        
        return await job_queue.start_workers(count)
    
    async def stop_workers(self, queue: str) -> None:
        """Stop workers for a queue."""
        job_queue = self._queues.get(queue)
        
        if job_queue:
            await job_queue.stop_workers()
    
    async def start_all(self) -> None:
        """Start workers for all queues."""
        for queue in self._queues.values():
            await queue.start_workers()
    
    async def stop_all(self) -> None:
        """Stop all workers."""
        for queue in self._queues.values():
            await queue.stop_workers()
    
    async def get_queue_stats(self, queue: str) -> Optional[QueueStats]:
        """Get queue statistics."""
        job_queue = self._queues.get(queue)
        
        if job_queue:
            return await job_queue.get_stats()
        
        return None
    
    async def get_stats(self) -> JobQueueStats:
        """Get overall statistics."""
        total_jobs = 0
        pending_jobs = 0
        running_jobs = 0
        total_workers = 0
        
        for queue in self._queues.values():
            stats = await queue.get_stats()
            pending_jobs += stats.pending_jobs
            running_jobs += stats.running_jobs
            total_workers += stats.workers_count
        
        return JobQueueStats(
            total_queues=len(self._queues),
            total_jobs=total_jobs,
            pending_jobs=pending_jobs,
            running_jobs=running_jobs,
            total_workers=total_workers,
        )
    
    def add_listener(self, listener: Callable) -> None:
        """Add event listener."""
        self._listeners.append(listener)


# Factory functions
def create_job_queue_manager() -> JobQueueManager:
    """Create job queue manager."""
    return JobQueueManager()


def create_job(
    queue: str,
    payload: Dict[str, Any],
    **kwargs,
) -> Job:
    """Create a job."""
    return Job(queue=queue, payload=payload, **kwargs)


def create_retry_policy(
    max_retries: int = 3,
    strategy: RetryStrategy = RetryStrategy.EXPONENTIAL,
    base_delay_seconds: float = 1.0,
) -> RetryPolicy:
    """Create retry policy."""
    return RetryPolicy(
        max_retries=max_retries,
        strategy=strategy,
        base_delay_seconds=base_delay_seconds,
    )


__all__ = [
    # Exceptions
    "JobError",
    "JobNotFoundError",
    "QueueNotFoundError",
    # Enums
    "JobStatus",
    "JobPriority",
    "RetryStrategy",
    # Data classes
    "RetryPolicy",
    "Job",
    "QueueConfig",
    "Worker",
    "QueueStats",
    "JobQueueStats",
    # Store
    "JobStore",
    "InMemoryJobStore",
    # Queue
    "JobQueue",
    # Manager
    "JobQueueManager",
    # Factory functions
    "create_job_queue_manager",
    "create_job",
    "create_retry_policy",
]



"""
Enterprise Saga Pattern Module.

Provides saga orchestration for distributed transactions,
compensation handling, and eventual consistency.

Example:
    # Define saga
    saga = Saga("order_saga")
    
    saga.step("reserve_inventory", reserve_inventory, compensate_inventory)
    saga.step("process_payment", process_payment, refund_payment)
    saga.step("ship_order", ship_order, cancel_shipment)
    
    # Execute saga
    result = await saga.execute(order_data)
    
    # With decorator
    @saga_step(name="reserve", compensate=compensate_reserve)
    async def reserve_inventory(context: SagaContext):
        ...
"""

from __future__ import annotations

import asyncio
import logging
import traceback
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from functools import wraps
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

logger = logging.getLogger(__name__)

T = TypeVar('T')


class SagaError(Exception):
    """Saga error."""
    pass


class SagaExecutionError(SagaError):
    """Saga execution failed."""
    
    def __init__(
        self,
        message: str,
        step_name: str,
        original_error: Exception,
    ):
        super().__init__(message)
        self.step_name = step_name
        self.original_error = original_error


class CompensationError(SagaError):
    """Compensation failed."""
    
    def __init__(
        self,
        message: str,
        failed_compensations: List[Tuple[str, Exception]],
    ):
        super().__init__(message)
        self.failed_compensations = failed_compensations


class SagaStatus(str, Enum):
    """Saga execution status."""
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    COMPENSATING = "compensating"
    COMPENSATED = "compensated"
    FAILED = "failed"
    PARTIALLY_COMPENSATED = "partially_compensated"


class StepStatus(str, Enum):
    """Step execution status."""
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    COMPENSATING = "compensating"
    COMPENSATED = "compensated"
    COMPENSATION_FAILED = "compensation_failed"
    SKIPPED = "skipped"


StepFunction = Callable[['SagaContext'], Awaitable[Any]]
CompensateFunction = Callable[['SagaContext'], Awaitable[None]]


@dataclass
class StepResult:
    """Result of step execution."""
    step_name: str
    status: StepStatus
    output: Any = None
    error: Optional[Exception] = None
    started_at: datetime = field(default_factory=datetime.now)
    completed_at: Optional[datetime] = None
    duration_ms: float = 0.0


@dataclass
class SagaResult:
    """Result of saga execution."""
    saga_id: str
    saga_name: str
    status: SagaStatus
    steps: List[StepResult] = field(default_factory=list)
    output: Any = None
    error: Optional[Exception] = None
    started_at: datetime = field(default_factory=datetime.now)
    completed_at: Optional[datetime] = None
    duration_ms: float = 0.0


@dataclass
class SagaContext:
    """
    Context passed between saga steps.
    """
    saga_id: str
    saga_name: str
    input_data: Any
    step_outputs: Dict[str, Any] = field(default_factory=dict)
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    def get_output(self, step_name: str) -> Any:
        """Get output from a previous step."""
        return self.step_outputs.get(step_name)
    
    def set_metadata(self, key: str, value: Any) -> None:
        """Set metadata value."""
        self.metadata[key] = value
    
    def get_metadata(self, key: str, default: Any = None) -> Any:
        """Get metadata value."""
        return self.metadata.get(key, default)


@dataclass
class SagaStep:
    """Definition of a saga step."""
    name: str
    execute: StepFunction
    compensate: Optional[CompensateFunction] = None
    timeout_seconds: float = 30.0
    retry_count: int = 0
    retry_delay_seconds: float = 1.0
    condition: Optional[Callable[[SagaContext], bool]] = None


class SagaStore(ABC):
    """Abstract saga state store."""
    
    @abstractmethod
    async def save(self, saga_id: str, state: Dict[str, Any]) -> None:
        """Save saga state."""
        pass
    
    @abstractmethod
    async def load(self, saga_id: str) -> Optional[Dict[str, Any]]:
        """Load saga state."""
        pass
    
    @abstractmethod
    async def delete(self, saga_id: str) -> bool:
        """Delete saga state."""
        pass


class InMemorySagaStore(SagaStore):
    """In-memory saga store."""
    
    def __init__(self):
        self._store: Dict[str, Dict[str, Any]] = {}
    
    async def save(self, saga_id: str, state: Dict[str, Any]) -> None:
        self._store[saga_id] = state
    
    async def load(self, saga_id: str) -> Optional[Dict[str, Any]]:
        return self._store.get(saga_id)
    
    async def delete(self, saga_id: str) -> bool:
        return self._store.pop(saga_id, None) is not None


class Saga:
    """
    Saga orchestrator for distributed transactions.
    
    Example:
        saga = Saga("order_saga")
        saga.step("reserve", reserve_fn, compensate_fn)
        saga.step("pay", pay_fn, refund_fn)
        
        result = await saga.execute(order_data)
    """
    
    def __init__(
        self,
        name: str,
        store: Optional[SagaStore] = None,
    ):
        self._name = name
        self._steps: List[SagaStep] = []
        self._store = store or InMemorySagaStore()
        self._hooks: Dict[str, List[Callable]] = {
            "before_step": [],
            "after_step": [],
            "on_compensate": [],
            "on_complete": [],
            "on_error": [],
        }
    
    @property
    def name(self) -> str:
        return self._name
    
    def step(
        self,
        name: str,
        execute: StepFunction,
        compensate: Optional[CompensateFunction] = None,
        timeout_seconds: float = 30.0,
        retry_count: int = 0,
        condition: Optional[Callable[[SagaContext], bool]] = None,
    ) -> 'Saga':
        """Add a step to the saga."""
        step = SagaStep(
            name=name,
            execute=execute,
            compensate=compensate,
            timeout_seconds=timeout_seconds,
            retry_count=retry_count,
            condition=condition,
        )
        self._steps.append(step)
        return self
    
    def on_before_step(
        self,
        callback: Callable[[str, SagaContext], Awaitable[None]],
    ) -> 'Saga':
        """Add before step hook."""
        self._hooks["before_step"].append(callback)
        return self
    
    def on_after_step(
        self,
        callback: Callable[[str, StepResult, SagaContext], Awaitable[None]],
    ) -> 'Saga':
        """Add after step hook."""
        self._hooks["after_step"].append(callback)
        return self
    
    def on_compensate(
        self,
        callback: Callable[[str, SagaContext], Awaitable[None]],
    ) -> 'Saga':
        """Add compensation hook."""
        self._hooks["on_compensate"].append(callback)
        return self
    
    async def execute(
        self,
        input_data: Any,
        saga_id: Optional[str] = None,
    ) -> SagaResult:
        """Execute the saga."""
        saga_id = saga_id or str(uuid.uuid4())
        start_time = datetime.now()
        
        context = SagaContext(
            saga_id=saga_id,
            saga_name=self._name,
            input_data=input_data,
        )
        
        result = SagaResult(
            saga_id=saga_id,
            saga_name=self._name,
            status=SagaStatus.RUNNING,
            started_at=start_time,
        )
        
        completed_steps: List[SagaStep] = []
        
        try:
            # Execute each step
            for step in self._steps:
                # Check condition
                if step.condition and not step.condition(context):
                    step_result = StepResult(
                        step_name=step.name,
                        status=StepStatus.SKIPPED,
                    )
                    result.steps.append(step_result)
                    continue
                
                # Execute step
                step_result = await self._execute_step(step, context)
                result.steps.append(step_result)
                
                if step_result.status == StepStatus.COMPLETED:
                    context.step_outputs[step.name] = step_result.output
                    completed_steps.append(step)
                else:
                    # Step failed, start compensation
                    raise SagaExecutionError(
                        f"Step {step.name} failed",
                        step.name,
                        step_result.error or Exception("Unknown error"),
                    )
            
            # All steps completed
            result.status = SagaStatus.COMPLETED
            result.output = context.step_outputs
            
            # Run completion hooks
            for hook in self._hooks["on_complete"]:
                await hook(result, context)
        
        except SagaExecutionError as e:
            logger.error(f"Saga {self._name} failed at step {e.step_name}")
            
            # Run error hooks
            for hook in self._hooks["on_error"]:
                await hook(e, context)
            
            # Compensate completed steps in reverse order
            compensation_result = await self._compensate(
                completed_steps,
                context,
            )
            
            result.status = compensation_result
            result.error = e
        
        # Finalize result
        result.completed_at = datetime.now()
        result.duration_ms = (
            result.completed_at - result.started_at
        ).total_seconds() * 1000
        
        # Persist final state
        await self._store.save(saga_id, {
            "status": result.status.value,
            "steps": [
                {
                    "name": s.step_name,
                    "status": s.status.value,
                }
                for s in result.steps
            ],
        })
        
        return result
    
    async def _execute_step(
        self,
        step: SagaStep,
        context: SagaContext,
    ) -> StepResult:
        """Execute a single step with retry."""
        start_time = datetime.now()
        
        result = StepResult(
            step_name=step.name,
            status=StepStatus.RUNNING,
            started_at=start_time,
        )
        
        # Run before hooks
        for hook in self._hooks["before_step"]:
            await hook(step.name, context)
        
        attempts = step.retry_count + 1
        last_error: Optional[Exception] = None
        
        for attempt in range(attempts):
            try:
                # Execute with timeout
                output = await asyncio.wait_for(
                    step.execute(context),
                    timeout=step.timeout_seconds,
                )
                
                result.status = StepStatus.COMPLETED
                result.output = output
                last_error = None
                break
            
            except asyncio.TimeoutError:
                last_error = asyncio.TimeoutError(
                    f"Step {step.name} timed out after {step.timeout_seconds}s"
                )
                logger.warning(
                    f"Step {step.name} timed out (attempt {attempt + 1}/{attempts})"
                )
            
            except Exception as e:
                last_error = e
                logger.warning(
                    f"Step {step.name} failed: {e} (attempt {attempt + 1}/{attempts})"
                )
            
            # Wait before retry
            if attempt < attempts - 1:
                await asyncio.sleep(step.retry_delay_seconds)
        
        if last_error:
            result.status = StepStatus.FAILED
            result.error = last_error
        
        # Finalize timing
        result.completed_at = datetime.now()
        result.duration_ms = (
            result.completed_at - result.started_at
        ).total_seconds() * 1000
        
        # Run after hooks
        for hook in self._hooks["after_step"]:
            await hook(step.name, result, context)
        
        return result
    
    async def _compensate(
        self,
        steps: List[SagaStep],
        context: SagaContext,
    ) -> SagaStatus:
        """Compensate completed steps in reverse order."""
        failed_compensations: List[Tuple[str, Exception]] = []
        
        # Reverse order compensation
        for step in reversed(steps):
            if not step.compensate:
                continue
            
            # Run compensation hooks
            for hook in self._hooks["on_compensate"]:
                await hook(step.name, context)
            
            try:
                await step.compensate(context)
                logger.info(f"Compensated step: {step.name}")
            
            except Exception as e:
                logger.error(f"Compensation failed for {step.name}: {e}")
                failed_compensations.append((step.name, e))
        
        if failed_compensations:
            return SagaStatus.PARTIALLY_COMPENSATED
        
        return SagaStatus.COMPENSATED


class SagaOrchestrator:
    """
    Orchestrates multiple sagas.
    """
    
    def __init__(
        self,
        store: Optional[SagaStore] = None,
    ):
        self._sagas: Dict[str, Saga] = {}
        self._store = store or InMemorySagaStore()
        self._running: Dict[str, asyncio.Task] = {}
    
    def register(self, saga: Saga) -> None:
        """Register a saga."""
        self._sagas[saga.name] = saga
    
    async def start(
        self,
        saga_name: str,
        input_data: Any,
    ) -> str:
        """Start a saga execution."""
        saga = self._sagas.get(saga_name)
        
        if not saga:
            raise SagaError(f"Saga not found: {saga_name}")
        
        saga_id = str(uuid.uuid4())
        
        task = asyncio.create_task(saga.execute(input_data, saga_id))
        self._running[saga_id] = task
        
        return saga_id
    
    async def wait(self, saga_id: str) -> SagaResult:
        """Wait for saga completion."""
        task = self._running.get(saga_id)
        
        if not task:
            raise SagaError(f"Saga not running: {saga_id}")
        
        return await task
    
    async def get_status(self, saga_id: str) -> Optional[SagaStatus]:
        """Get saga status."""
        state = await self._store.load(saga_id)
        
        if state:
            return SagaStatus(state["status"])
        
        return None
    
    async def cancel(self, saga_id: str) -> bool:
        """Cancel a running saga."""
        task = self._running.get(saga_id)
        
        if task and not task.done():
            task.cancel()
            del self._running[saga_id]
            return True
        
        return False


class SagaBuilder:
    """
    Fluent builder for sagas.
    """
    
    def __init__(self, name: str):
        self._name = name
        self._steps: List[SagaStep] = []
        self._store: Optional[SagaStore] = None
    
    def with_store(self, store: SagaStore) -> 'SagaBuilder':
        """Set saga store."""
        self._store = store
        return self
    
    def step(
        self,
        name: str,
        execute: StepFunction,
    ) -> 'SagaStepBuilder':
        """Start building a step."""
        return SagaStepBuilder(self, name, execute)
    
    def _add_step(self, step: SagaStep) -> None:
        """Add step (internal)."""
        self._steps.append(step)
    
    def build(self) -> Saga:
        """Build the saga."""
        saga = Saga(self._name, self._store)
        
        for step in self._steps:
            saga._steps.append(step)
        
        return saga


class SagaStepBuilder:
    """Builder for saga steps."""
    
    def __init__(
        self,
        parent: SagaBuilder,
        name: str,
        execute: StepFunction,
    ):
        self._parent = parent
        self._name = name
        self._execute = execute
        self._compensate: Optional[CompensateFunction] = None
        self._timeout = 30.0
        self._retry_count = 0
        self._condition: Optional[Callable] = None
    
    def compensate(
        self,
        compensate_fn: CompensateFunction,
    ) -> 'SagaStepBuilder':
        """Set compensation function."""
        self._compensate = compensate_fn
        return self
    
    def timeout(self, seconds: float) -> 'SagaStepBuilder':
        """Set timeout."""
        self._timeout = seconds
        return self
    
    def retry(self, count: int) -> 'SagaStepBuilder':
        """Set retry count."""
        self._retry_count = count
        return self
    
    def when(
        self,
        condition: Callable[[SagaContext], bool],
    ) -> 'SagaStepBuilder':
        """Set execution condition."""
        self._condition = condition
        return self
    
    def next(
        self,
        name: str,
        execute: StepFunction,
    ) -> 'SagaStepBuilder':
        """Add next step."""
        self._finalize()
        return SagaStepBuilder(self._parent, name, execute)
    
    def build(self) -> Saga:
        """Build the saga."""
        self._finalize()
        return self._parent.build()
    
    def _finalize(self) -> None:
        """Finalize current step."""
        step = SagaStep(
            name=self._name,
            execute=self._execute,
            compensate=self._compensate,
            timeout_seconds=self._timeout,
            retry_count=self._retry_count,
            condition=self._condition,
        )
        self._parent._add_step(step)


# Decorators
def saga_step(
    name: str,
    compensate: Optional[CompensateFunction] = None,
    timeout: float = 30.0,
    retry: int = 0,
) -> Callable:
    """
    Decorator to define a saga step.
    
    Example:
        @saga_step("reserve_inventory", compensate=release_inventory)
        async def reserve_inventory(context: SagaContext):
            ...
    """
    def decorator(func: StepFunction) -> SagaStep:
        return SagaStep(
            name=name,
            execute=func,
            compensate=compensate,
            timeout_seconds=timeout,
            retry_count=retry,
        )
    
    return decorator


def compensates(step_name: str) -> Callable:
    """
    Decorator to mark a function as compensation.
    
    Example:
        @compensates("reserve_inventory")
        async def release_inventory(context: SagaContext):
            ...
    """
    def decorator(func: CompensateFunction) -> CompensateFunction:
        func._compensates_step = step_name
        return func
    
    return decorator


def transactional_saga(
    name: str,
    store: Optional[SagaStore] = None,
) -> Callable:
    """
    Decorator to wrap a function in a saga.
    
    Example:
        @transactional_saga("process_order")
        async def process_order(order: dict):
            yield step("validate", validate_order)
            yield step("reserve", reserve_inventory, release_inventory)
            yield step("charge", charge_payment, refund_payment)
    """
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> SagaResult:
            saga = Saga(name, store)
            
            # Get steps from generator
            gen = func(*args, **kwargs)
            
            for step_def in gen:
                if isinstance(step_def, SagaStep):
                    saga._steps.append(step_def)
                elif isinstance(step_def, tuple):
                    saga.step(*step_def)
            
            # Execute with first arg as input
            input_data = args[0] if args else kwargs
            return await saga.execute(input_data)
        
        return wrapper
    
    return decorator


# Factory functions
def create_saga(
    name: str,
    store: Optional[SagaStore] = None,
) -> Saga:
    """Create a saga."""
    return Saga(name, store)


def create_saga_builder(name: str) -> SagaBuilder:
    """Create a saga builder."""
    return SagaBuilder(name)


def create_saga_orchestrator(
    store: Optional[SagaStore] = None,
) -> SagaOrchestrator:
    """Create a saga orchestrator."""
    return SagaOrchestrator(store)


__all__ = [
    # Exceptions
    "SagaError",
    "SagaExecutionError",
    "CompensationError",
    # Enums
    "SagaStatus",
    "StepStatus",
    # Data classes
    "StepResult",
    "SagaResult",
    "SagaContext",
    "SagaStep",
    # Core classes
    "SagaStore",
    "InMemorySagaStore",
    "Saga",
    "SagaOrchestrator",
    "SagaBuilder",
    "SagaStepBuilder",
    # Decorators
    "saga_step",
    "compensates",
    "transactional_saga",
    # Factory functions
    "create_saga",
    "create_saga_builder",
    "create_saga_orchestrator",
]



from __future__ import annotations

import logging
import time
from collections import deque
from typing import Any, Callable, Dict, List

from .exceptions import CriterionEvaluationError  # noqa: F401 - exported for library users

logger = logging.getLogger(__name__)

_MAX_RESULTS = 5_000


class EvaluationSystem:
    """Core evaluation system with bounded result history."""

    def __init__(self, max_results: int = _MAX_RESULTS) -> None:
        self.criteria: Dict[str, Callable[[Any], bool]] = {}
        self.results: deque[Dict[str, Any]] = deque(maxlen=max_results)

    def define_criterion(self, name: str, evaluation_fn: Callable[[Any], bool]) -> None:
        self.criteria[name] = evaluation_fn
        logger.info("Defined evaluation criterion '%s'", name)

    def evaluate(self, data: Any) -> Dict[str, bool]:
        evaluation_result: Dict[str, bool] = {}
        for name, fn in self.criteria.items():
            try:
                evaluation_result[name] = fn(data)
            except (TypeError, ValueError, KeyError, AttributeError) as e:
                evaluation_result[name] = False
                logger.warning("Criterion '%s' evaluation failed: %s", name, e)
            except Exception as e:  # noqa: BLE001 - Fail safe for unknown errors
                evaluation_result[name] = False
                logger.exception("Unexpected error in criterion '%s'", name)
        self.results.append({"data": data, "result": evaluation_result, "timestamp": time.time()})
        return evaluation_result

    def get_results(self) -> List[Dict[str, Any]]:
        return list(self.results)



"""
Enterprise Calendar Service Module.

Scheduling, recurring events, timezone handling,
availability, and calendar management.

Example:
    # Create calendar service
    calendar = create_calendar_service()
    
    # Create event
    event = await calendar.create_event(
        title="Team Meeting",
        start=datetime(2024, 1, 15, 10, 0),
        end=datetime(2024, 1, 15, 11, 0),
        timezone="America/New_York",
        recurrence="FREQ=WEEKLY;BYDAY=MO,WE,FR",
    )
    
    # Get events for date range
    events = await calendar.get_events(
        start=datetime(2024, 1, 1),
        end=datetime(2024, 1, 31),
    )
    
    # Check availability
    available = await calendar.check_availability(
        user_id="user_123",
        start=datetime(2024, 1, 15, 14, 0),
        duration=60,  # minutes
    )
"""

from __future__ import annotations

import asyncio
import json
import logging
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timedelta, date, time
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo

T = TypeVar('T')

logger = logging.getLogger(__name__)


class CalendarError(Exception):
    """Calendar error."""
    pass


class EventNotFoundError(CalendarError):
    """Event not found."""
    pass


class ConflictError(CalendarError):
    """Scheduling conflict."""
    pass


class EventStatus(str, Enum):
    """Event status."""
    CONFIRMED = "confirmed"
    TENTATIVE = "tentative"
    CANCELLED = "cancelled"


class RecurrenceFrequency(str, Enum):
    """Recurrence frequency."""
    DAILY = "DAILY"
    WEEKLY = "WEEKLY"
    MONTHLY = "MONTHLY"
    YEARLY = "YEARLY"


class Weekday(str, Enum):
    """Weekday."""
    MONDAY = "MO"
    TUESDAY = "TU"
    WEDNESDAY = "WE"
    THURSDAY = "TH"
    FRIDAY = "FR"
    SATURDAY = "SA"
    SUNDAY = "SU"


@dataclass
class Attendee:
    """Event attendee."""
    email: str = ""
    name: str = ""
    status: str = "needs_action"  # accepted, declined, tentative, needs_action
    required: bool = True


@dataclass
class Reminder:
    """Event reminder."""
    minutes_before: int = 15
    method: str = "popup"  # popup, email, sms


@dataclass
class RecurrenceRule:
    """Recurrence rule."""
    frequency: RecurrenceFrequency = RecurrenceFrequency.WEEKLY
    interval: int = 1
    count: Optional[int] = None
    until: Optional[datetime] = None
    by_day: List[Weekday] = field(default_factory=list)
    by_month_day: List[int] = field(default_factory=list)
    by_month: List[int] = field(default_factory=list)
    
    def to_rrule_string(self) -> str:
        """Convert to RRULE string."""
        parts = [f"FREQ={self.frequency.value}"]
        
        if self.interval != 1:
            parts.append(f"INTERVAL={self.interval}")
        
        if self.count:
            parts.append(f"COUNT={self.count}")
        
        if self.until:
            parts.append(f"UNTIL={self.until.strftime('%Y%m%dT%H%M%SZ')}")
        
        if self.by_day:
            parts.append(f"BYDAY={','.join(d.value for d in self.by_day)}")
        
        if self.by_month_day:
            parts.append(f"BYMONTHDAY={','.join(str(d) for d in self.by_month_day)}")
        
        if self.by_month:
            parts.append(f"BYMONTH={','.join(str(m) for m in self.by_month)}")
        
        return ";".join(parts)
    
    @classmethod
    def from_rrule_string(cls, rrule: str) -> "RecurrenceRule":
        """Parse from RRULE string."""
        rule = cls()
        
        for part in rrule.split(";"):
            if "=" not in part:
                continue
            
            key, value = part.split("=", 1)
            
            if key == "FREQ":
                rule.frequency = RecurrenceFrequency(value)
            elif key == "INTERVAL":
                rule.interval = int(value)
            elif key == "COUNT":
                rule.count = int(value)
            elif key == "UNTIL":
                rule.until = datetime.strptime(value, "%Y%m%dT%H%M%SZ")
            elif key == "BYDAY":
                rule.by_day = [Weekday(d) for d in value.split(",")]
            elif key == "BYMONTHDAY":
                rule.by_month_day = [int(d) for d in value.split(",")]
            elif key == "BYMONTH":
                rule.by_month = [int(m) for m in value.split(",")]
        
        return rule


@dataclass
class Event:
    """Calendar event."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    calendar_id: str = ""
    title: str = ""
    description: str = ""
    location: str = ""
    start: datetime = field(default_factory=datetime.utcnow)
    end: datetime = field(default_factory=datetime.utcnow)
    all_day: bool = False
    timezone: str = "UTC"
    status: EventStatus = EventStatus.CONFIRMED
    attendees: List[Attendee] = field(default_factory=list)
    reminders: List[Reminder] = field(default_factory=list)
    recurrence: Optional[RecurrenceRule] = None
    recurrence_id: Optional[str] = None
    color: str = ""
    metadata: Dict[str, Any] = field(default_factory=dict)
    created_by: str = ""
    created_at: datetime = field(default_factory=datetime.utcnow)
    updated_at: datetime = field(default_factory=datetime.utcnow)
    
    @property
    def duration(self) -> timedelta:
        """Get event duration."""
        return self.end - self.start
    
    @property
    def duration_minutes(self) -> int:
        """Get duration in minutes."""
        return int(self.duration.total_seconds() / 60)
    
    def overlaps(self, other: "Event") -> bool:
        """Check if overlaps with another event."""
        return self.start < other.end and self.end > other.start


@dataclass
class Calendar:
    """Calendar."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    description: str = ""
    owner_id: str = ""
    color: str = "#4285F4"
    timezone: str = "UTC"
    is_primary: bool = False
    is_shared: bool = False
    shared_with: List[str] = field(default_factory=list)
    metadata: Dict[str, Any] = field(default_factory=dict)
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class TimeSlot:
    """Time slot."""
    start: datetime = field(default_factory=datetime.utcnow)
    end: datetime = field(default_factory=datetime.utcnow)
    available: bool = True
    event_id: Optional[str] = None


@dataclass
class Availability:
    """User availability."""
    user_id: str = ""
    slots: List[TimeSlot] = field(default_factory=list)
    working_hours: Dict[str, Tuple[time, time]] = field(default_factory=dict)
    timezone: str = "UTC"


@dataclass
class CalendarStats:
    """Calendar statistics."""
    total_calendars: int = 0
    total_events: int = 0
    recurring_events: int = 0
    events_this_week: int = 0


# Event store
class EventStore(ABC):
    """Event storage."""
    
    @abstractmethod
    async def save(self, event: Event) -> None:
        """Save event."""
        pass
    
    @abstractmethod
    async def get(self, event_id: str) -> Optional[Event]:
        """Get event."""
        pass
    
    @abstractmethod
    async def query(
        self,
        calendar_id: Optional[str] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> List[Event]:
        """Query events."""
        pass
    
    @abstractmethod
    async def delete(self, event_id: str) -> bool:
        """Delete event."""
        pass


class InMemoryEventStore(EventStore):
    """In-memory event store."""
    
    def __init__(self):
        self._events: Dict[str, Event] = {}
    
    async def save(self, event: Event) -> None:
        event.updated_at = datetime.utcnow()
        self._events[event.id] = event
    
    async def get(self, event_id: str) -> Optional[Event]:
        return self._events.get(event_id)
    
    async def query(
        self,
        calendar_id: Optional[str] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> List[Event]:
        results = []
        
        for event in self._events.values():
            if calendar_id and event.calendar_id != calendar_id:
                continue
            if start and event.end < start:
                continue
            if end and event.start > end:
                continue
            
            results.append(event)
        
        return sorted(results, key=lambda e: e.start)
    
    async def delete(self, event_id: str) -> bool:
        return self._events.pop(event_id, None) is not None


# Calendar store
class CalendarStore(ABC):
    """Calendar storage."""
    
    @abstractmethod
    async def save(self, calendar: Calendar) -> None:
        """Save calendar."""
        pass
    
    @abstractmethod
    async def get(self, calendar_id: str) -> Optional[Calendar]:
        """Get calendar."""
        pass
    
    @abstractmethod
    async def list(self, owner_id: Optional[str] = None) -> List[Calendar]:
        """List calendars."""
        pass
    
    @abstractmethod
    async def delete(self, calendar_id: str) -> bool:
        """Delete calendar."""
        pass


class InMemoryCalendarStore(CalendarStore):
    """In-memory calendar store."""
    
    def __init__(self):
        self._calendars: Dict[str, Calendar] = {}
    
    async def save(self, calendar: Calendar) -> None:
        self._calendars[calendar.id] = calendar
    
    async def get(self, calendar_id: str) -> Optional[Calendar]:
        return self._calendars.get(calendar_id)
    
    async def list(self, owner_id: Optional[str] = None) -> List[Calendar]:
        calendars = list(self._calendars.values())
        if owner_id:
            calendars = [c for c in calendars if c.owner_id == owner_id]
        return calendars
    
    async def delete(self, calendar_id: str) -> bool:
        return self._calendars.pop(calendar_id, None) is not None


# Calendar service
class CalendarService:
    """Calendar service."""
    
    def __init__(
        self,
        event_store: Optional[EventStore] = None,
        calendar_store: Optional[CalendarStore] = None,
        default_timezone: str = "UTC",
    ):
        self._events = event_store or InMemoryEventStore()
        self._calendars = calendar_store or InMemoryCalendarStore()
        self._default_timezone = default_timezone
        self._stats = CalendarStats()
    
    # Calendar management
    async def create_calendar(
        self,
        name: str,
        owner_id: str,
        description: str = "",
        timezone: Optional[str] = None,
        **kwargs,
    ) -> Calendar:
        """Create calendar."""
        calendar = Calendar(
            name=name,
            owner_id=owner_id,
            description=description,
            timezone=timezone or self._default_timezone,
            **kwargs,
        )
        await self._calendars.save(calendar)
        self._stats.total_calendars += 1
        
        logger.info(f"Calendar created: {name}")
        
        return calendar
    
    async def get_calendar(self, calendar_id: str) -> Optional[Calendar]:
        """Get calendar."""
        return await self._calendars.get(calendar_id)
    
    async def list_calendars(
        self,
        owner_id: Optional[str] = None,
    ) -> List[Calendar]:
        """List calendars."""
        return await self._calendars.list(owner_id)
    
    async def delete_calendar(self, calendar_id: str) -> bool:
        """Delete calendar."""
        result = await self._calendars.delete(calendar_id)
        if result:
            self._stats.total_calendars -= 1
        return result
    
    # Event management
    async def create_event(
        self,
        title: str,
        start: datetime,
        end: Optional[datetime] = None,
        duration: Optional[int] = None,  # minutes
        calendar_id: str = "",
        timezone: Optional[str] = None,
        recurrence: Optional[str] = None,
        attendees: Optional[List[Dict[str, str]]] = None,
        **kwargs,
    ) -> Event:
        """Create event."""
        if end is None:
            if duration:
                end = start + timedelta(minutes=duration)
            else:
                end = start + timedelta(hours=1)
        
        event = Event(
            calendar_id=calendar_id,
            title=title,
            start=start,
            end=end,
            timezone=timezone or self._default_timezone,
            attendees=[
                Attendee(**a) for a in (attendees or [])
            ],
            **kwargs,
        )
        
        # Parse recurrence
        if recurrence:
            event.recurrence = RecurrenceRule.from_rrule_string(recurrence)
            self._stats.recurring_events += 1
        
        await self._events.save(event)
        self._stats.total_events += 1
        
        logger.info(f"Event created: {title}")
        
        return event
    
    async def get_event(self, event_id: str) -> Optional[Event]:
        """Get event."""
        return await self._events.get(event_id)
    
    async def update_event(
        self,
        event_id: str,
        **updates,
    ) -> Optional[Event]:
        """Update event."""
        event = await self._events.get(event_id)
        if not event:
            return None
        
        for key, value in updates.items():
            if hasattr(event, key):
                setattr(event, key, value)
        
        await self._events.save(event)
        
        return event
    
    async def delete_event(
        self,
        event_id: str,
        delete_series: bool = False,
    ) -> bool:
        """Delete event."""
        event = await self._events.get(event_id)
        if not event:
            return False
        
        if delete_series and event.recurrence:
            # Delete all instances
            events = await self._events.query(calendar_id=event.calendar_id)
            for e in events:
                if e.recurrence_id == event_id or e.id == event_id:
                    await self._events.delete(e.id)
        else:
            await self._events.delete(event_id)
        
        self._stats.total_events -= 1
        return True
    
    async def get_events(
        self,
        calendar_id: Optional[str] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        expand_recurring: bool = True,
    ) -> List[Event]:
        """Get events."""
        events = await self._events.query(
            calendar_id=calendar_id,
            start=start,
            end=end,
        )
        
        if expand_recurring and start and end:
            expanded = []
            for event in events:
                if event.recurrence:
                    instances = self._expand_recurring(event, start, end)
                    expanded.extend(instances)
                else:
                    expanded.append(event)
            events = expanded
        
        return sorted(events, key=lambda e: e.start)
    
    def _expand_recurring(
        self,
        event: Event,
        range_start: datetime,
        range_end: datetime,
    ) -> List[Event]:
        """Expand recurring event."""
        instances: List[Event] = []
        
        if not event.recurrence:
            return [event]
        
        rule = event.recurrence
        current = event.start
        count = 0
        max_instances = rule.count or 365
        
        while current < range_end and count < max_instances:
            if current >= range_start:
                instance = Event(
                    id=f"{event.id}_{current.isoformat()}",
                    calendar_id=event.calendar_id,
                    title=event.title,
                    description=event.description,
                    location=event.location,
                    start=current,
                    end=current + event.duration,
                    timezone=event.timezone,
                    status=event.status,
                    attendees=event.attendees,
                    recurrence_id=event.id,
                    color=event.color,
                    metadata=event.metadata,
                )
                instances.append(instance)
            
            # Advance to next occurrence
            if rule.frequency == RecurrenceFrequency.DAILY:
                current += timedelta(days=rule.interval)
            elif rule.frequency == RecurrenceFrequency.WEEKLY:
                current += timedelta(weeks=rule.interval)
            elif rule.frequency == RecurrenceFrequency.MONTHLY:
                # Simple month increment
                month = current.month + rule.interval
                year = current.year + (month - 1) // 12
                month = ((month - 1) % 12) + 1
                try:
                    current = current.replace(year=year, month=month)
                except ValueError:
                    # Handle end of month
                    current = current.replace(year=year, month=month + 1, day=1)
            elif rule.frequency == RecurrenceFrequency.YEARLY:
                current = current.replace(year=current.year + rule.interval)
            
            count += 1
            
            if rule.until and current > rule.until:
                break
        
        return instances
    
    # Availability
    async def check_availability(
        self,
        user_id: str,
        start: datetime,
        duration: int,  # minutes
        calendar_id: Optional[str] = None,
    ) -> bool:
        """Check if time slot is available."""
        end = start + timedelta(minutes=duration)
        
        events = await self.get_events(
            calendar_id=calendar_id,
            start=start,
            end=end,
        )
        
        # Filter by user
        user_events = [
            e for e in events
            if e.created_by == user_id or any(
                a.email == user_id for a in e.attendees
            )
        ]
        
        return len(user_events) == 0
    
    async def find_available_slots(
        self,
        user_id: str,
        date: date,
        duration: int,  # minutes
        working_hours: Tuple[time, time] = (time(9, 0), time(17, 0)),
        calendar_id: Optional[str] = None,
    ) -> List[TimeSlot]:
        """Find available time slots."""
        start = datetime.combine(date, working_hours[0])
        end = datetime.combine(date, working_hours[1])
        
        events = await self.get_events(
            calendar_id=calendar_id,
            start=start,
            end=end,
        )
        
        # Find gaps
        slots: List[TimeSlot] = []
        current = start
        
        for event in sorted(events, key=lambda e: e.start):
            if event.start > current:
                # There's a gap
                gap_minutes = int((event.start - current).total_seconds() / 60)
                if gap_minutes >= duration:
                    slots.append(TimeSlot(
                        start=current,
                        end=event.start,
                        available=True,
                    ))
            current = max(current, event.end)
        
        # Check remaining time
        if current < end:
            gap_minutes = int((end - current).total_seconds() / 60)
            if gap_minutes >= duration:
                slots.append(TimeSlot(
                    start=current,
                    end=end,
                    available=True,
                ))
        
        return slots
    
    # Timezone handling
    def convert_timezone(
        self,
        dt: datetime,
        from_tz: str,
        to_tz: str,
    ) -> datetime:
        """Convert datetime between timezones."""
        from_zone = ZoneInfo(from_tz)
        to_zone = ZoneInfo(to_tz)
        
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=from_zone)
        
        return dt.astimezone(to_zone)
    
    def now(self, timezone: Optional[str] = None) -> datetime:
        """Get current time in timezone."""
        tz = ZoneInfo(timezone or self._default_timezone)
        return datetime.now(tz)
    
    # Stats
    def get_stats(self) -> CalendarStats:
        """Get statistics."""
        return self._stats


# Factory functions
def create_calendar_service(
    default_timezone: str = "UTC",
) -> CalendarService:
    """Create calendar service."""
    return CalendarService(default_timezone=default_timezone)


def create_calendar(
    name: str,
    owner_id: str = "",
    **kwargs,
) -> Calendar:
    """Create calendar."""
    return Calendar(name=name, owner_id=owner_id, **kwargs)


def create_event(
    title: str,
    start: datetime,
    end: Optional[datetime] = None,
    **kwargs,
) -> Event:
    """Create event."""
    return Event(
        title=title,
        start=start,
        end=end or (start + timedelta(hours=1)),
        **kwargs,
    )


def create_recurrence(
    frequency: RecurrenceFrequency = RecurrenceFrequency.WEEKLY,
    interval: int = 1,
    **kwargs,
) -> RecurrenceRule:
    """Create recurrence rule."""
    return RecurrenceRule(
        frequency=frequency,
        interval=interval,
        **kwargs,
    )


__all__ = [
    # Exceptions
    "CalendarError",
    "EventNotFoundError",
    "ConflictError",
    # Enums
    "EventStatus",
    "RecurrenceFrequency",
    "Weekday",
    # Data classes
    "Attendee",
    "Reminder",
    "RecurrenceRule",
    "Event",
    "Calendar",
    "TimeSlot",
    "Availability",
    "CalendarStats",
    # Stores
    "EventStore",
    "InMemoryEventStore",
    "CalendarStore",
    "InMemoryCalendarStore",
    # Service
    "CalendarService",
    # Factory functions
    "create_calendar_service",
    "create_calendar",
    "create_event",
    "create_recurrence",
]
